Substitution = dict["VariableArgument", "Argument"]


class Argument:
    def __init__(self, name: str):
        self.name: str = name

    def __str__(self) -> str:
        return self.name

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.name))

    def __eq__(self, other) -> bool:
        return type(self) is type(other) and self.name == other.name

    def is_variable(self) -> bool:
        return False

    def is_constant(self) -> bool:
        return False

    def apply(self, sub: Substitution) -> "Argument":
        return self


class VariableArgument(Argument):
    def is_variable(self) -> bool:
        return True

    def apply(self, sub: Substitution) -> "VariableArgument":
        return sub.get(self, self)


class ObjectArgument(Argument):
    def is_constant(self) -> bool:
        return True


class ArgumentDefinition:
    def __init__(self, name: str, type_name: str | None):
        self.name: str = name
        self.type_name: str = type_name

    def __str__(self) -> str:
        if self.type_name is not None:
            return f"{self.name} - {self.type_name}"
        return self.name



from plado.parser.tokenizer import Category, tokenize


def test_singleton():
    tokens = list(tokenize("("))
    assert len(tokens) == 1
    assert tokens[0].cat == Category.LBRACK


def test_pair():
    tokens = list(tokenize("()"))
    assert len(tokens) == 2
    assert tokens[0].cat == Category.LBRACK
    assert tokens[1].cat == Category.RBRACK


def test_pair_ws():
    tokens = list(tokenize("(    )"))
    assert len(tokens) == 2
    assert tokens[0].cat == Category.LBRACK
    assert tokens[1].cat == Category.RBRACK


def test_pair_trailing_ws():
    tokens = list(tokenize("   (    )   "))
    assert len(tokens) == 2
    assert tokens[0].cat == Category.LBRACK
    assert tokens[1].cat == Category.RBRACK


def test_pair_newline_start():
    tokens = list(tokenize("""
()    """))
    assert len(tokens) == 2
    assert tokens[0].cat == Category.LBRACK
    assert tokens[1].cat == Category.RBRACK


def test_pair_newline_end():
    tokens = list(tokenize("""
()    
  """))
    assert len(tokens) == 2
    assert tokens[0].cat == Category.LBRACK
    assert tokens[1].cat == Category.RBRACK


def test_action():
    action_pddl = """
    (:action put-on
        :parameters (?x ?y - object)
        :precondition (and (clear ?x) (holding ?y))
        :effect (and (not (clear ?x)) (not (holding ?y)) (on ?x ?y))
        )
    """
    tokens = list(tokenize(action_pddl))
    expected = [
        Category.LBRACK,
        Category.ACTION,
        Category.NAME,
        Category.PARAMETERS,
        Category.LBRACK,
        Category.VARIABLE,
        Category.VARIABLE,
        Category.MINUS,
        Category.NAME,
        Category.RBRACK,
        Category.PRECONDITION,
        Category.LBRACK,
        Category.AND,
        Category.LBRACK,
        Category.NAME,
        Category.VARIABLE,
        Category.RBRACK,
        Category.LBRACK,
        Category.NAME,
        Category.VARIABLE,
        Category.RBRACK,
        Category.RBRACK,
        Category.EFFECT,
        Category.LBRACK,
        Category.AND,
        Category.LBRACK,
        Category.NOT,
        Category.LBRACK,
        Category.NAME,
        Category.VARIABLE,
        Category.RBRACK,
        Category.RBRACK,
        Category.LBRACK,
        Category.NOT,
        Category.LBRACK,
        Category.NAME,
        Category.VARIABLE,
        Category.RBRACK,
        Category.RBRACK,
        Category.LBRACK,
        Category.NAME,
        Category.VARIABLE,
        Category.VARIABLE,
        Category.RBRACK,
        Category.RBRACK,
        Category.RBRACK,
    ]
    assert len(tokens) == len(expected)
    for i, cat in enumerate(expected):
        assert tokens[i].cat == cat


def test_comment():
    pddl = "(define (domain ;this is a comment ))"
    expected = [Category.LBRACK, Category.DEFINE, Category.LBRACK, Category.DOMAIN_DEF]
    tokens = list(tokenize(pddl))
    assert len(tokens) == len(expected)
    for i, tok in enumerate(expected):
        assert tokens[i].cat == tok


def test_comment_multi_line():
    pddl = """(define;this is a comment 
   (domain 
   ;; comment
   ))"""
    expected = [
        Category.LBRACK,
        Category.DEFINE,
        Category.LBRACK,
        Category.DOMAIN_DEF,
        Category.RBRACK,
        Category.RBRACK,
    ]
    tokens = list(tokenize(pddl))
    assert len(tokens) == len(expected)
    for i, tok in enumerate(expected):
        assert tokens[i].cat == tok



from plado.pddl.action import *
from plado.pddl.arguments import *
from plado.pddl.boolean_expression import *
from plado.pddl.derived_predicate import *
from plado.pddl.domain import *
from plado.pddl.effects import *
from plado.pddl.metrics import *
from plado.pddl.numeric_expression import *
from plado.pddl.problem import *
from plado.pddl.type import *



import itertools
from collections.abc import Callable, Iterable, Mapping

from plado.datalog import numeric
from plado.utils import UnionFind


class IntRef:
    def __init__(self, n: int):
        self.n: int = n


class Constant:
    def __init__(self, idx: int, variable: bool):
        self.id: int = idx
        self.variable: bool = variable

    def __eq__(self, o) -> bool:
        return type(o) is type(self) and o.id == self.id and self.variable == o.variable

    def __str__(self) -> str:
        return ("?x" if self.variable else "o") + str(self.id)

    def is_variable(self) -> bool:
        return self.variable


class Atom:
    def __init__(self, relation_id: int, arguments: Iterable[Constant]):
        self.relation_id: int = relation_id
        self.arguments: tuple[Constant] = tuple(arguments)

    def __eq__(self, o) -> bool:
        return (
            type(self) is type(o)
            and self.relation_id == o.relation_id
            and self.arguments == o.arguments
        )

    def __str__(self) -> str:
        return (
            f"(P{self.relation_id}"
            f"{' ' if len(self.arguments) > 0 else ''}"
            f"{' '.join((str(arg) for arg in self.arguments))})"
        )

    def __repr__(self) -> str:
        return str(self)

    def get_variables(self) -> list[int]:
        return list((arg.id for arg in self.arguments if arg.is_variable()))

    def standardize_arguments(
        self,
        next_var_id: IntRef,
        notify_variable: Callable[[int, int], None],
        notify_object: Callable[[int, int], None],
    ) -> "Atom":
        """
        After standardization, every argument is a unique variable. Might create
        fresh variables to satisfy this property. Calls notify_variable for
        variables newly introduced substituing duplicate occurences of the same
        variable, and notify_object for variables substituing objects, in both
        cases passing the old argument as first and the new variable as second argument.
        """
        arguments = []
        in_use: set[int] = set()
        for arg in self.arguments:
            if arg.is_variable():
                if arg.id in in_use:
                    arguments.append(Constant(next_var_id.n, True))
                    notify_variable(arg.id, next_var_id.n)
                    next_var_id.n += 1
                else:
                    in_use.add(arg.id)
                    arguments.append(arg)
            else:
                arguments.append(Constant(next_var_id.n, True))
                notify_object(arg.id, next_var_id.n)
                next_var_id.n += 1

        return Atom(self.relation_id, arguments)

    def substitute(self, sub: Mapping[int, int]) -> "Atom":
        """
        Substitutes all variables according to the given substitution
        mapping.
        """
        return Atom(
            self.relation_id,
            (
                (Constant(sub[arg.id], True) if arg.is_variable() else arg)
                for arg in self.arguments
            ),
        )


class Clause:
    def __init__(
        self,
        head: Atom,
        pos_body: Iterable[Atom],
        neg_body: Iterable[Atom],
        constraints: Iterable[numeric.NumericConstraint],
    ):
        self.head: Atom = head
        self.pos_body: tuple[Atom] = tuple(pos_body)
        self.neg_body: tuple[Atom] = tuple(neg_body)
        self.constraints: tuple[numeric.NumericConstraint] = tuple(constraints)

    def __str__(self) -> str:
        res = [str(self.head), "=:"]
        res.extend((str(a) for a in self.pos_body))
        res.extend((f"not {a}" for a in self.neg_body))
        res.extend((str(c) for c in self.constraints))
        return " ".join(res)

    def is_trivial(self) -> bool:
        """
        A clause is trivial if its body is empty.
        """
        return len(self.pos_body) + len(self.neg_body) + len(self.constraints) == 0

    def _make_0_indexed(self, num_variables: IntRef) -> "Clause":
        """
        Replaces variable ids, making them range from [0, ..., num_variables)
        """
        ids = sorted(
            set(
                itertools.chain(
                    itertools.chain.from_iterable((
                        (arg.id for arg in atom.arguments if arg.is_variable())
                        for atom in itertools.chain(
                            [self.head], self.pos_body, self.neg_body
                        )
                    )),
                    itertools.chain.from_iterable(
                        (c.expr.get_variables() for c in self.constraints)
                    ),
                )
            )
        )
        id_map: dict[int, int] = {old: pos for pos, old in enumerate(ids)}
        num_variables.n = len(ids)
        return Clause(
            self.head.substitute(id_map),
            (a.substitute(id_map) for a in self.pos_body),
            (a.substitute(id_map) for a in self.neg_body),
            (c.substitute(id_map) for c in self.constraints),
        )

    def _merge_variables(self, eq_relation: int, num_variables: int) -> "Clause":
        """
        Finds each equality atom in the body containing two
        variables. Replaces the two variables by a unique representative.
        """

        def is_var_eq(atom: Atom) -> bool:
            return (
                atom.relation_id == eq_relation
                and atom.arguments[0].is_variable()
                and atom.arguments[1].is_variable()
            )

        eq_atoms = False
        uf = UnionFind(num_variables)
        for atom in self.pos_body:
            if is_var_eq(atom):
                eq_atoms = True
                x = atom.arguments[0].id
                y = atom.arguments[1].id
                uf.merge(x, y)
        if not eq_atoms:
            return self
        new_id = [None for _ in range(num_variables)]
        j = 0
        for i in range(num_variables):
            if uf[i] == i:
                new_id[i] = j
                j += 1
            else:
                assert new_id[uf[i]] is not None
                new_id[i] = new_id[uf[i]]
        return Clause(
            self.head.substitute(new_id),
            (a.substitute(new_id) for a in self.pos_body if not is_var_eq(a)),
            (a.substitute(new_id) for a in self.neg_body),
            (c.substitute(new_id) for c in self.constraints),
        )

    def standardize_variables(self, eq_relation: int) -> "Clause":
        """
        Makes sure all atoms in the body have only variables as arguments,
        and that no variable appears twice in any atom. Introduces additional
        variables and equality atoms to satisfy these properties.
        """
        num_variables = IntRef(0)
        clause = (
            self._make_0_indexed(num_variables)
            ._merge_variables(eq_relation, num_variables.n)
            ._make_0_indexed(num_variables)
        )
        next_var_id = IntRef(num_variables.n)
        pos_body = []
        neg_body = []

        def on_variable(old_var: int, new_var: int) -> None:
            pos_body.append(
                Atom(eq_relation, [Constant(old_var, True), Constant(new_var, True)])
            )

        def on_object(obj_id: int, new_var: int) -> None:
            pos_body.append(
                Atom(eq_relation, [Constant(new_var, True), Constant(obj_id, False)])
            )

        assert clause.head.relation_id != eq_relation

        def standardize(atom, atoms):
            if atom.relation_id == eq_relation:
                assert len(atom.arguments) == 2
                assert (
                    atom.arguments[0].is_variable() or atom.arguments[1].is_variable()
                )
                if atom.arguments[0] != atom.arguments[1]:
                    atoms.append(atom)
            else:
                atoms.append(
                    atom.standardize_arguments(next_var_id, on_variable, on_object)
                )

        new_head = clause.head.standardize_arguments(
            next_var_id, on_variable, on_object
        )

        for atom in clause.pos_body:
            standardize(atom, pos_body)

        for atom in clause.neg_body:
            standardize(atom, neg_body)

        return Clause(
            new_head,
            pos_body,
            neg_body,
            self.constraints,
        )


class DatalogProgram:
    def __init__(self):
        self.relation_arities: list[int] = []
        self.clauses: list[Clause] = []
        self.trivial_clauses: list[Atom] = []
        self.equality_relation: int | None = None

    def num_relations(self) -> int:
        return len(self.relation_arities)

    def add_relation(self, arity: int) -> int:
        self.relation_arities.append(arity)
        return len(self.relation_arities) - 1

    def add_clause(self, clause: Clause) -> None:
        assert self.equality_relation is not None
        if clause.is_trivial():
            self.trivial_clauses.append(clause.head)
        else:
            self.clauses.append(clause.standardize_variables(self.equality_relation))



import itertools
from collections.abc import Iterable
from typing import Optional

from skdecide import (
    DeterministicPlanningDomain,
    DiscreteDistribution,
    GoalMDPDomain,
    ImplicitSpace,
    Space,
    Value,
)
from skdecide.hub.space.gym import ListSpace

from plado.parser import parse_and_normalize
from plado.semantics.applicable_actions_generator import ApplicableActionsGenerator
from plado.semantics.goal_checker import GoalChecker
from plado.semantics.successor_generator import SuccessorGenerator
from plado.semantics.task import State as PyddlState
from plado.semantics.task import Task
from plado.utils import Float

Action = tuple[int, tuple[int]]


class State:
    def __init__(self, state: PyddlState, cost_function: set[int] = set()):
        self.atoms: tuple[tuple[tuple[int]]] = tuple(
            tuple(sorted(state.atoms[p])) for p in range(len(state.atoms))
        )
        self.fluents: tuple[tuple[tuple[int], int]] = tuple(
            (
                tuple(
                    (args, int(state.fluents[f][args]))
                    for args in sorted(state.fluents[f].keys())
                )
                if f not in cost_function
                else tuple()
            )
            for f in range(len(state.fluents))
        )

    def to_plado(self, cost_functions: set[int]) -> PyddlState:
        state = PyddlState(0, len(self.fluents))
        state.atoms = self.atoms
        for f, values in enumerate(self.fluents):
            for args, val in values:
                state.fluents[f][args] = Float(val)
        for f in cost_functions:
            state.fluents[f][tuple()] = Float(0)
        return state

    def __hash__(self):
        return hash((self.atoms, self.fluents))

    def __eq__(self, o) -> bool:
        return (
            isinstance(o, State) and self.atoms == o.atoms and self.fluents == o.fluents
        )


class D:
    T_state = State  # Type of states
    T_observation = T_state  # Type of observations
    T_event = Action  # Type of events
    T_value = float  # Type of transition values (rewards or costs)
    T_predicate = bool  # Type of test results
    T_info = None  # Type of additional information in environment outcome


class ObservationSpace(Space[D.T_observation]):
    def __init__(
        self,
        predicate_arities: Iterable[int],
        function_arities: Iterable[int],
        num_objects: int,
    ):
        self.predicate_arities: tuple[int] = tuple(predicate_arities)
        self.function_arities: tuple[int] = tuple(function_arities)
        self.num_objects: int = num_objects

    def contains(self, x: D.T_observation) -> bool:
        if len(x.atoms) != len(self.predicate_arities) or len(x.fluents) != len(
            self.function_arities
        ):
            return False
        for p in range(x.atoms):
            for params in x.atoms[p]:
                if (
                    len(params) != self.predicate_arities[p]
                    or min(params) < 0
                    or max(params) >= self.num_objects
                ):
                    return False
        for f in range(x.fluents):
            for params, _ in x.fluents[f]:
                if (
                    len(params) != self.function_arities[f]
                    or min(params) < 0
                    or max(params) >= self.num_objects
                ):
                    return False
        return True


class ActionSpace(Space[D.T_event]):
    def __init__(self, action_arities: Iterable[int], num_objects: int):
        self.action_arities: tuple[int] = tuple(action_arities)
        self.num_objects: int = num_objects

    def contains(self, a: D.T_event) -> bool:
        return (
            a[0] >= 0
            and a[0] < len(self.action_arities)
            and len(a[1]) == self.action_arities[a[0]]
            and min(a[1]) >= 0
            and max(a[1]) < self.num_objects
        )


class SkdPyddlBaseDomain(D):

    def __init__(self, domain_path: str, problem_path: str):
        self.domain_path: str = domain_path
        self.problem_path: str = problem_path
        domain, problem = parse_and_normalize(domain_path, problem_path)
        self.task: Task = Task(domain, problem)
        self.check_goal: GoalChecker = GoalChecker(self.task)
        self.aops_gen: ApplicableActionsGenerator = ApplicableActionsGenerator(
            self.task
        )
        self.succ_gen: SuccessorGenerator = SuccessorGenerator(self.task)
        self.total_cost: int | None = None
        for i, f in enumerate(self.task.functions):
            if f.name == "total-cost":
                self.total_cost = i
                break
        self.cost_functions: set[int] = set(
            [self.total_cost] if self.total_cost is not None else []
        )
        self.transition_cost: dict[tuple[State, Action, State], int] = {}

    def _translate_state(self, state: PyddlState) -> State:
        return State(state, self.cost_functions)

    def _get_cost_from_state(self, state: PyddlState) -> int:
        if self.total_cost is None:
            return 1  # assume unit cost
        return int(state.fluents[self.total_cost][tuple()])

    def _get_initial_state_(self) -> D.T_state:
        return self._translate_state(self.task.initial_state)

    def _get_observation_space_(self) -> Space[D.T_observation]:
        return ObservationSpace(
            (
                len(self.task.predicates[p].parameters)
                for p in range(self.task.num_fluent_predicates)
            ),
            (len(f.parameters) for f in self.task.functions),
            len(self.task.objects),
        )

    def _get_action_space_(self) -> Space[D.T_event]:
        return ActionSpace(
            (a.parameters for a in self.task.actions), len(self.task.objects)
        )

    def _get_goals_(self) -> Space[D.T_observation]:
        return ImplicitSpace(lambda s: self.check_goal(s.to_plado(self.cost_functions)))

    def _is_terminal(self, state: D.T_state) -> D.T_predicate:
        return self.check_goal(state.to_plado(self.cost_functions))

    def _get_applicable_actions_from(self, memory: D.T_state) -> Space[D.T_event]:
        return ListSpace(list(self.aops_gen(memory.to_plado(self.cost_functions))))

    def _get_next_state_distribution(
        self, memory: D.T_state, action: D.T_event
    ) -> DiscreteDistribution[D.T_state]:
        successors = self.succ_gen(memory.to_plado(self.cost_functions), action)
        ts = [(self._translate_state(succ), float(prob)) for succ, prob in successors]
        for i in range(len(ts)):
            c = self._get_cost_from_state(successors[i][0])
            if c != 1:
                self.transition_cost[(memory, action, ts[i])] = c
        return DiscreteDistribution(ts)

    def _get_next_state(self, memory: D.T_state, action: D.T_event) -> D.T_state:
        successors = self.succ_gen(memory.to_plado(self.cost_functions), action)
        successor = successors[0][0]
        t = self._translate_state(successor)
        c = self._get_cost_from_state(successor)
        if c != 1:
            self.transition_cost[(memory, action, t)] = c
        return t

    def _get_transition_value(
        self,
        memory: D.T_state,
        action: D.T_event,
        next_state: Optional[D.T_state] = None,
    ) -> Value[D.T_value]:
        return Value(cost=self.transition_cost.get((memory, action, next_state), 1))


class SkdPyddlDomain(SkdPyddlBaseDomain, DeterministicPlanningDomain):
    pass


class SkdPPyddlDomain(SkdPyddlBaseDomain, GoalMDPDomain):
    pass



import itertools

import pytest

from plado import pddl
from plado.semantics import task as T
from plado.semantics.applicable_actions_generator import ApplicableActionsGenerator
from plado.semantics.successor_generator import SuccessorGenerator
from plado.semantics.task import Task
from plado.utils import is_equal


@pytest.fixture
def pddl_domain_1():
    p = pddl.Predicate("p", (pddl.ArgumentDefinition("?x", "object"),))
    q = pddl.Predicate("q", (pddl.ArgumentDefinition("?x", "object"),))

    precondition = p((pddl.VariableArgument("?x"),))
    effect = pddl.AtomEffect("q", (pddl.VariableArgument("?x"),))
    action = pddl.Action(
        "a", (pddl.ArgumentDefinition("?x", "object"),), precondition, effect
    )

    return pddl.Domain("domain", [], [], [], [p, q], [], [action], [])


@pytest.fixture
def pddl_problem_1_1():
    n = 10
    objs = [pddl.ArgumentDefinition(f"o{i}", "object") for i in range(n)]
    init = [pddl.Atom("p", (pddl.ObjectArgument(f"o{i}"),)) for i in range(n)]
    goal = pddl.Conjunction(
        [pddl.Atom("q", (pddl.ObjectArgument(f"o{i}"),)) for i in range(n)]
    )
    return pddl.Problem("problem", "domain", objs, init, goal, None, None)


def test_normalization_1(pddl_domain_1, pddl_problem_1_1):
    task = Task(pddl_domain_1, pddl_problem_1_1)
    assert len(task.initial_state.atoms) == 1
    # p
    assert task.initial_state.atoms[0] == set()
    # q and =
    assert len(task.static_facts) == 2
    # q
    assert task.static_facts[0] == set(((i,) for i in range(10)))
    # =
    assert task.static_facts[1] == set(((i, i) for i in range(10)))
    assert len(task.actions) == 1
    a = task.actions[0]
    assert len(a.precondition.atoms) == 1
    assert a.precondition.atoms[0].predicate == 1
    assert len(a.precondition.atoms[0].args) == 1
    assert a.precondition.atoms[0].args[0] == None
    assert len(a.precondition.get_variables()) == 1


def test_aops_gen_1(pddl_domain_1, pddl_problem_1_1):
    task = Task(pddl_domain_1, pddl_problem_1_1)
    gen = ApplicableActionsGenerator(task)
    aops = set(gen(task.initial_state))
    assert aops == set(((0, (i,)) for i in range(10)))


def test_successor_generator_1(pddl_domain_1, pddl_problem_1_1):
    task = Task(pddl_domain_1, pddl_problem_1_1)
    succ_gen = SuccessorGenerator(task)
    for i in range(10):
        successors = succ_gen(task.initial_state, (0, (i,)))
        assert len(successors) == 1
        assert (i,) in successors[0][0].atoms[0]
        assert successors[0][1] == 1.0


@pytest.fixture
def pddl_domain_2():
    p = pddl.Predicate("p", (pddl.ArgumentDefinition("?x", "object"),))
    q = pddl.Predicate("q", (pddl.ArgumentDefinition("?x", "object"),))
    r = pddl.Predicate("r", (pddl.ArgumentDefinition("?x", "object"),))
    s = pddl.Predicate(
        "s",
        (
            pddl.ArgumentDefinition("?x", "object"),
            pddl.ArgumentDefinition("?y", "object"),
        ),
    )

    precondition = pddl.Conjunction([
        s((
            pddl.VariableArgument("?x"),
            pddl.VariableArgument("?y"),
        )),
        s((
            pddl.VariableArgument("?y"),
            pddl.VariableArgument("?z"),
        )),
    ])
    effect = pddl.ConjunctiveEffect([
        pddl.ProbabilisticEffect([
            pddl.ProbabilisticOutcome(
                pddl.NumericConstant("0.5"),
                pddl.ConjunctiveEffect([
                    pddl.ConditionalEffect(
                        p((pddl.VariableArgument("?z"),)),
                        pddl.AtomEffect("q", (pddl.VariableArgument("?z"),)),
                    ),
                    pddl.ConditionalEffect(
                        p((pddl.VariableArgument("?z"),)),
                        pddl.NegativeEffect(
                            pddl.AtomEffect("p", (pddl.VariableArgument("?z"),))
                        ),
                    ),
                ]),
            )
        ]),
        pddl.ProbabilisticEffect([
            pddl.ProbabilisticOutcome(
                pddl.NumericConstant("0.5"),
                pddl.ConjunctiveEffect([
                    pddl.ConditionalEffect(
                        p((pddl.VariableArgument("?z"),)),
                        pddl.AtomEffect("r", (pddl.VariableArgument("?z"),)),
                    ),
                    pddl.ConditionalEffect(
                        p((pddl.VariableArgument("?z"),)),
                        pddl.NegativeEffect(
                            pddl.AtomEffect("p", (pddl.VariableArgument("?z"),))
                        ),
                    ),
                ]),
            )
        ]),
    ])
    action = pddl.Action(
        "a",
        tuple((pddl.ArgumentDefinition(x, "object") for x in ["?x", "?y", "?z"])),
        precondition,
        effect,
    )

    return pddl.Domain("domain", [], [], [], [p, q, r, s], [], [action], [])


@pytest.fixture
def pddl_problem_2_1():
    objs = [pddl.ArgumentDefinition(f"o{i}", "object") for i in range(6)]
    init = [pddl.Atom("p", (pddl.ObjectArgument(f"o{i}"),)) for i in range(6)] + [
        pddl.Atom("s", (pddl.ObjectArgument(f"o{i}"), pddl.ObjectArgument(f"o{i+1}")))
        for i in range(5)
        if i != 2
    ]
    goal = pddl.Conjunction(
        [pddl.Atom("q", (pddl.ObjectArgument(f"o{i}"),)) for i in [2, 5]]
    )
    return pddl.Problem("problem", "domain", objs, init, goal, None, None)


def test_normalization_2(pddl_domain_2, pddl_problem_2_1):
    task = Task(pddl_domain_2, pddl_problem_2_1)
    assert len(task.initial_state.atoms) == 3  # p, q, r
    assert len(task.static_facts) == 2  # s, =
    assert task.initial_state.atoms[0] == set(((i,) for i in range(6)))
    assert task.initial_state.atoms[1] == set()
    assert task.initial_state.atoms[2] == set()
    assert task.static_facts[0] == set(((i, i + 1) for i in range(5) if i != 2))
    assert len(task.actions) == 1
    a = task.actions[0]
    assert len(a.precondition.atoms) == 2
    for i in (0, 1):
        assert a.precondition.atoms[i].predicate == 3
        assert len(a.precondition.atoms[i].args) == 2
        assert a.precondition.atoms[i].args[0] is None
        assert a.precondition.atoms[i].args[1] is None
        assert len(a.precondition.get_variables()) == 3
    assert len(a.effect.effects) == 2
    for i in [0, 1]:
        assert len(a.effect.effects[i].outcomes) == 1


def test_aops_gen_2(pddl_domain_2, pddl_problem_2_1):
    task = Task(pddl_domain_2, pddl_problem_2_1)
    gen = ApplicableActionsGenerator(task)
    aops = set(gen(task.initial_state))
    assert aops == set(((0, (i, i + 1, i + 2)) for i in [0, 3]))


def test_successor_generator_2(pddl_domain_2, pddl_problem_2_1):
    task = Task(pddl_domain_2, pddl_problem_2_1)
    succ_gen = SuccessorGenerator(task)
    state = task.initial_state
    for i in [0, 3]:
        successors = succ_gen(state, (0, (i, i + 1, i + 2)))
        assert len(successors) == 4
        for succ, prob in successors:
            assert (len(succ.atoms[1]), len(succ.atoms[2])) in (
                (0, 0),
                (1, 0),
                (0, 1),
                (1, 1),
            )
            assert is_equal(prob, 0.25)
            assert len(succ.atoms[1]) + len(succ.atoms[2]) == 0 or succ.atoms[0] == set(
                (j,) for j in range(6) if i + 2 != j
            )
            assert len(succ.atoms[1]) == 0 or succ.atoms[1] == set([(i + 2,)])
            assert len(succ.atoms[2]) == 0 or succ.atoms[2] == set([(i + 2,)])


@pytest.fixture
def pddl_domain_3():
    p = pddl.Predicate("p", (pddl.ArgumentDefinition("?x", "object"),))
    nz = pddl.Predicate("nz", (pddl.ArgumentDefinition("?x", "object"),))
    r = pddl.Predicate(
        "r",
        (
            pddl.ArgumentDefinition("?x", "object"),
            pddl.ArgumentDefinition("?y", "object"),
            pddl.ArgumentDefinition("?z", "object"),
        ),
    )
    s = pddl.Predicate(
        "s",
        (
            pddl.ArgumentDefinition("?x", "object"),
            pddl.ArgumentDefinition("?y", "object"),
        ),
    )

    eq = pddl.Predicate(
        "=",
        [
            pddl.ArgumentDefinition("?x", "object"),
            pddl.ArgumentDefinition("?y", "object"),
        ],
    )
    predicates = [p, nz, r, s, eq]

    # x + y = z if y + x = z
    pred0 = r((
        pddl.VariableArgument("?y"),
        pddl.VariableArgument("?x"),
        pddl.VariableArgument("?z"),
    ))

    # x + y = z if x + yy = zz where yy + 1 = y and zz + 1 = zz
    pred1 = pddl.Exists(
        [
            pddl.ArgumentDefinition("?yy", "object"),
            pddl.ArgumentDefinition("?zz", "object"),
        ],
        pddl.Conjunction([
            s((pddl.VariableArgument("?yy"), pddl.VariableArgument("?y"))),
            s((pddl.VariableArgument("?zz"), pddl.VariableArgument("?z"))),
            r((
                pddl.VariableArgument("?x"),
                pddl.VariableArgument("?yy"),
                pddl.VariableArgument("?zz"),
            )),
        ]),
    )

    # x + 0 = x
    pred2 = pddl.Conjunction([
        pddl.Atom("=", [pddl.VariableArgument("?x"), pddl.VariableArgument("?z")]),
        pddl.Negation(pddl.Atom("nz", [pddl.VariableArgument("?y")])),
    ])

    # not zero -> has predecessor
    pred3 = pddl.Exists(
        [pddl.ArgumentDefinition("?xx", "object")],
        pddl.Atom("s", [pddl.VariableArgument("?xx"), pddl.VariableArgument("?x")]),
    )

    derived_predicates = [
        pddl.DerivedPredicate(nz, pred3),
        pddl.DerivedPredicate(r, pred0),
        pddl.DerivedPredicate(r, pred1),
        pddl.DerivedPredicate(r, pred2),
    ]

    precondition = pddl.Conjunction([
        p((pddl.VariableArgument("?x"),)),
        r((
            pddl.VariableArgument("?z"),
            pddl.VariableArgument("?y"),
            pddl.VariableArgument("?x"),
        )),
    ])
    effect = pddl.ConjunctiveEffect([
        pddl.ProbabilisticEffect([
            pddl.ProbabilisticOutcome(
                pddl.NumericConstant("0.5"),
                pddl.ConjunctiveEffect([
                    pddl.NegativeEffect(
                        pddl.AtomEffect("p", [pddl.VariableArgument("?x")])
                    ),
                ]),
            ),
            pddl.ProbabilisticOutcome(
                pddl.NumericConstant("0.5"),
                pddl.ConjunctiveEffect([
                    pddl.NegativeEffect(
                        pddl.AtomEffect("p", [pddl.VariableArgument("?x")])
                    ),
                    pddl.AtomEffect("p", [pddl.VariableArgument("?z")]),
                ]),
            ),
        ]),
    ])
    decrease = pddl.Action(
        "decrease",
        tuple((pddl.ArgumentDefinition(x, "object") for x in ["?x", "?y", "?z"])),
        precondition,
        effect,
    )

    precondition = pddl.Conjunction([
        p((pddl.VariableArgument("?x"),)),
        r((
            pddl.VariableArgument("?x"),
            pddl.VariableArgument("?y"),
            pddl.VariableArgument("?z"),
        )),
    ])
    effect = pddl.ConjunctiveEffect([
        pddl.ProbabilisticEffect([
            pddl.ProbabilisticOutcome(
                pddl.NumericConstant("0.5"),
                pddl.ConjunctiveEffect([
                    pddl.NegativeEffect(
                        pddl.AtomEffect("p", [pddl.VariableArgument("?x")])
                    ),
                ]),
            ),
            pddl.ProbabilisticOutcome(
                pddl.NumericConstant("0.5"),
                pddl.ConjunctiveEffect([
                    pddl.NegativeEffect(
                        pddl.AtomEffect("p", [pddl.VariableArgument("?x")])
                    ),
                    pddl.AtomEffect("p", [pddl.VariableArgument("?z")]),
                ]),
            ),
        ]),
    ])
    increase = pddl.Action(
        "increase",
        tuple((pddl.ArgumentDefinition(x, "object") for x in ["?x", "?y", "?z"])),
        precondition,
        effect,
    )

    return pddl.Domain(
        "domain", [], [], [], predicates, [], [increase, decrease], derived_predicates
    )


@pytest.fixture
def pddl_problem_3_1():
    objs = [pddl.ArgumentDefinition(f"o{i}", "object") for i in range(6)]
    init = (
        [pddl.Atom("p", (pddl.ObjectArgument(f"o{i}"),)) for i in [0]]
        + [
            pddl.Atom(
                "s", (pddl.ObjectArgument(f"o{i}"), pddl.ObjectArgument(f"o{i+1}"))
            )
            for i in range(5)
        ]
        + [
            pddl.Atom("=", [pddl.ObjectArgument(f"o{i}"), pddl.ObjectArgument(f"o{i}")])
            for i in range(5)
        ]
    )
    goal = pddl.Conjunction(
        [pddl.Atom("p", (pddl.ObjectArgument(f"o{i}"),)) for i in [5]]
    )
    return pddl.Problem("problem", "domain", objs, init, goal, None, None)


def test_normalization_3(pddl_domain_3, pddl_problem_3_1):
    task = Task(pddl_domain_3, pddl_problem_3_1)
    assert task.num_fluent_predicates == 1
    assert task.num_derived_predicates == 2
    assert task.num_static_predicates == 2
    assert task.initial_state.atoms[0] == set(((i,) for i in [0]))
    assert task.static_facts[0] == set(((i, i + 1) for i in range(5)))
    assert len(task.actions) == 2
    for action in task.actions:
        assert len(action.precondition.atoms) == 2
        assert action.precondition.atoms[0].predicate == 0
        assert len(action.precondition.atoms[0].variables) == 1
        assert len(action.precondition.atoms[1].variables) == 3
        assert len(action.effect.effects) == 1
        assert len(action.effect.effects[0].outcomes) == 2


def test_aops_gen_3(pddl_domain_3, pddl_problem_3_1):
    task = Task(pddl_domain_3, pddl_problem_3_1)
    gen = ApplicableActionsGenerator(task)
    aops = set(gen(task.initial_state))
    assert aops == set(
        itertools.chain(
            ((0, (0, i, i)) for i in range(6)),
            [(1, (0, 0, 0))],
        )
    )


def test_aops_gen_3_2(pddl_domain_3, pddl_problem_3_1):
    task = Task(pddl_domain_3, pddl_problem_3_1)
    gen = ApplicableActionsGenerator(task)
    task.initial_state.atoms[0] = set([(2,)])
    aops = set(gen(task.initial_state))
    assert aops == set(
        itertools.chain(
            ((1, (2, i, 2 - i)) for i in range(3)),
            ((0, (2, i, i + 2)) for i in range(4)),
        )
    )


def test_successor_generator_3(pddl_domain_3, pddl_problem_3_1):
    task = Task(pddl_domain_3, pddl_problem_3_1)
    gen = SuccessorGenerator(task)
    state = task.initial_state
    succs = gen(state, (0, (0, 3, 3)))
    assert len(succs) == 2
    assert len(succs[0][0].atoms[0]) == 0 or len(succs[1][0].atoms) == 0
    assert succs[0][0].atoms[0] == set([(3,)]) or succs[1][0].atoms[0] == set([(3,)])
    assert is_equal(succs[0][1], 0.5) and is_equal(succs[1][1], 0.5)



import argparse
import time

from plado.parser import parse_and_normalize
from plado.semantics.grounder import Grounder
from plado.semantics.task import Task


def main():
    p = argparse.ArgumentParser()
    p.add_argument("domain")
    p.add_argument("problem")
    args = p.parse_args()

    t = time.time()

    print("parsing pddl...")
    domain, problem = parse_and_normalize(args.domain, args.problem)

    print(f"normalizing pddl... [t={time.time() - t}s]")
    task = Task(domain, problem)

    print(f"starting grounding... [t={time.time() -t}s]")

    grounder = Grounder(task)

    facts = grounder.get_ground_facts()
    actions = grounder.get_ground_actions()

    print(
        f"grounded task into {len(facts)} facts and {len(actions)} actions [t={time.time() -t}s]"
    )

    print("============== facts ==============")
    for fact in facts:
        if fact[0] < task.num_fluent_predicates:
            print(task.dump_fact(fact[0], fact[1]))

    print()
    print("============== action ==============")
    for action in actions:
        print(task.dump_action(*action))


if __name__ == "__main__":
    main()



from plado import pddl


def visit_all_conditions_in_effect(
    eff: pddl.ActionEffect, visitor: pddl.BooleanExpressionVisitor
):
    class EffectVisitor(pddl.RecursiveActionEffectVisitor):
        def visit_atomic(self, effect: pddl.ActionEffect) -> bool:
            return False

        def visit_conditional_effect(self, effect: pddl.ConditionalEffect) -> bool:
            if effect.condition.traverse(visitor):
                return True
            return super().visit_conditional_effect(effect)

    eff_visitor = EffectVisitor()
    return eff.traverse(eff_visitor)


def visit_all_conditions(domain: pddl.Domain, visitor: pddl.BooleanExpressionVisitor):
    for action in domain.actions:
        action.precondition.traverse(visitor)
        visit_all_conditions_in_effect(action.effect, visitor)

    for predicate in domain.derived_predicates:
        predicate.condition.traverse(visitor)


def visit_all_expressions_in_condition(
    cond: pddl.BooleanExpression, visitor: pddl.NumericExpressionVisitor
):
    class ConditionVisitor(pddl.RecursiveBooleanExpressionVisitor):
        def visit_atomic(self, formula: pddl.BooleanExpression) -> bool:
            return False

        def visit_numeric_condition(self, cond: pddl.NumericComparison) -> bool:
            return cond.lhs.traverse(visitor) or cond.rhs.traverse(visitor)

    cv = ConditionVisitor()
    return cond.traverse(cv)


def visit_all_expressions_in_effect(
    eff: pddl.ActionEffect, visitor: pddl.NumericExpressionVisitor
):
    class EffectVisitor(pddl.RecursiveActionEffectVisitor):
        def visit_atomic(self, effect: pddl.ActionEffect) -> bool:
            return False

        def visit_numeric_effect(self, effect: pddl.NumericEffect) -> bool:
            return effect.variable.traverse(visitor) or effect.expression.traverse(
                visitor
            )

        def visit_probabilistic_effect(self, effect: pddl.ProbabilisticEffect) -> bool:
            for outcome in effect.outcomes:
                if outcome.probability.traverse(visitor):
                    return True
            return super().visit_probabilistic_effect(effect)

        def visit_conditional_effect(self, effect: pddl.ConditionalEffect) -> bool:
            if visit_all_expressions_in_condition(effect.condition, visitor):
                return True
            return super().visit_conditional_effect(effect)

    ev = EffectVisitor()
    return eff.traverse(ev)


def visit_all_expressions(domain: pddl.Domain, visitor: pddl.NumericExpressionVisitor):
    for action in domain.actions:
        visit_all_expressions_in_condition(action.precondition, visitor)
        visit_all_expressions_in_effect(action.effect, visitor)

    for predicate in domain.derived_predicates:
        visit_all_expressions_in_condition(predicate.condition, visitor)


def visit_all_effects(actions: tuple[pddl.Action], visitor: pddl.ActionEffectVisitor):
    for action in actions:
        action.effect.traverse(visitor)



from collections.abc import Iterable

from plado.pddl.action import Action
from plado.pddl.arguments import ArgumentDefinition
from plado.pddl.boolean_expression import Predicate
from plado.pddl.derived_predicate import DerivedPredicate
from plado.pddl.numeric_expression import Function
from plado.pddl.type import Type

CR = "\n"


class Domain:
    def __init__(
        self,
        name: str,
        requirements: Iterable[str],
        types: Iterable[Type],
        constants: Iterable[ArgumentDefinition],
        predicates: Iterable[Predicate],
        functions: Iterable[Function],
        actions: Iterable[Action],
        derived_predicates: Iterable[DerivedPredicate],
    ):
        self.name: str = name
        self.requirements: tuple[str] = requirements
        self.types: tuple[Type] = tuple(types)
        self.constants: tuple[ArgumentDefinition] = tuple(constants)
        self.predicates: tuple[Predicate] = tuple(predicates)
        self.functions: tuple[Function] = tuple(functions)
        self.actions: tuple[Action] = tuple(actions)
        self.derived_predicates: tuple[DerivedPredicate] = tuple(derived_predicates)

    def dump(self) -> str:
        result = [
            f"(define (domain {self.name})",
        ]
        if len(self.requirements) > 0:
            result.append(f"(:requirements {' '.join(self.requirements)})")
        if len(self.types) > 0:
            result.append("(:types")
            result.append(CR.join((f"  {str(t)}" for t in self.types)))
            result.append(")")
        if len(self.constants) > 0:
            result.append(f"(:constants {' '.join((str(o) for o in self.constants))})")
        if len(self.predicates) > 0:
            result.append("(:predicates")
            result.append(CR.join((f"  {str(p)}" for p in self.predicates)))
            result.append(")")
        if len(self.functions) > 0:
            result.append("(:functions")
            result.append(CR.join((f"  {str(p)}" for p in self.functions)))
            result.append(")")
        result.append(CR.join((p.dump(0) for p in self.derived_predicates)))
        result.append(CR.join((a.dump(0) for a in self.actions)))
        result.append(")")
        return CR.join(result)



import itertools

import pytest

from plado.datalog import evaluator
from plado.datalog.program import Atom, Clause, Constant


def Var(x: int) -> Constant:
    return Constant(x, True)


def Obj(x: int) -> Constant:
    return Constant(x, False)


def test_clause_reindexing():
    head = Atom(0, [Var(2), Var(1)])
    body = [
        Atom(0, [Var(1), Var(2)]),
    ]
    clause = Clause(head, body, [], [])
    normalized = clause.standardize_variables(1)
    assert normalized.head == Atom(0, [Var(1), Var(0)])
    assert len(normalized.pos_body) == 1
    assert len(normalized.neg_body) == 0
    assert len(normalized.constraints) == 0
    assert Atom(0, [Var(0), Var(1)]) in normalized.pos_body


def test_clause_standardization():
    head = Atom(1, [Var(2), Var(2)])
    body = [Atom(2, [Var(1), Var(2)])]
    neg_body = [Atom(3, [Var(1), Var(1)])]
    clause = Clause(head, body, neg_body, [])
    normalized = clause.standardize_variables(0)
    assert normalized.head == Atom(1, [Var(1), Var(2)])
    assert len(normalized.pos_body) == 3
    assert Atom(2, [Var(0), Var(1)]) in normalized.pos_body
    assert Atom(0, [Var(1), Var(2)]) in normalized.pos_body
    assert Atom(0, [Var(0), Var(3)]) in normalized.pos_body
    assert len(normalized.neg_body) == 1
    assert Atom(3, [Var(0), Var(3)]) in normalized.neg_body
    assert len(normalized.constraints) == 0


def test_clause_objects():
    head = Atom(1, [Var(0), Var(2)])
    body = [Atom(2, [Var(2), Obj(2)])]
    neg_body = []
    clause = Clause(head, body, neg_body, [])
    normalized = clause.standardize_variables(0)
    assert normalized.head == Atom(1, [Var(0), Var(1)])
    assert len(normalized.pos_body) == 2
    assert Atom(2, [Var(1), Var(2)]) in normalized.pos_body
    assert Atom(0, [Var(2), Obj(2)]) in normalized.pos_body
    assert len(normalized.neg_body) == 0
    assert len(normalized.constraints) == 0


def test_clause_normalization():
    head = Atom(1, [Var(0), Var(1)])
    eq = Atom(0, [Var(0), Var(1)])
    oeq = Atom(0, [Var(0), Obj(0)])
    body = [head, eq, oeq]
    neg_body = [head, eq, oeq]
    clause = Clause(head, body, neg_body, [])
    normalized = evaluator._normalize_clause(clause, 0, 2)
    assert normalized.head == head
    assert len(normalized.positive) == 1 and head in normalized.positive
    assert len(normalized.negative) == 1 and head in normalized.negative
    assert len(normalized.vars_eq) == 1 and (0, 1) in normalized.vars_eq
    assert len(normalized.vars_neq) == 1 and (0, 1) in normalized.vars_neq
    assert len(normalized.obj_eq) == 1 and (0, 0) in normalized.obj_eq
    assert len(normalized.obj_neq) == 1 and (0, 0) in normalized.obj_neq
    assert len(normalized.constrants) == 0
    assert normalized.num_variables == 2


@pytest.fixture
def AcycDepClauses() -> list[Clause]:
    return [
        Clause(
            Atom(1, [Var(0), Var(1)]),
            [Atom(2, [Var(1), Var(2)]), Atom(3, [Var(0), Var(2)])],
            [],
            [],
        ),
        Clause(Atom(2, [Var(0), Var(1)]), [Atom(3, [Var(0), Var(1)])], [], []),
        Clause(Atom(3, [Var(0), Var(1)]), [Atom(3, [Var(1), Var(0)])], [], []),
    ]


def test_dependency_graph_acyclic(AcycDepClauses):
    dg = evaluator._get_dependency_graph(
        5, [evaluator._normalize_clause(c, 0, 4) for c in AcycDepClauses]
    )
    assert len(dg) == 5
    assert set(dg[0]) == set()
    assert set(dg[1]) == set([2, 3])
    assert set(dg[2]) == set([3])
    assert set(dg[3]) == set([3])


def test_dependent_components_acyclic(AcycDepClauses):
    dg = evaluator._get_dependency_graph(
        5, [evaluator._normalize_clause(c, 0, 4) for c in AcycDepClauses]
    )
    components = evaluator._get_dependent_components(dg)
    assert len(components) == 5
    assert set(components[0]) == set([0])
    assert set(components[1]) == set([3])
    assert set(components[2]) == set([2])
    assert set(components[3]) == set([1])


def test_is_stratified(AcycDepClauses):
    clauses = [evaluator._normalize_clause(c, 0, 4) for c in AcycDepClauses]
    dg = evaluator._get_dependency_graph(5, clauses)
    components = evaluator._get_dependent_components(dg)
    assert evaluator._is_stratified(5, clauses, components)


@pytest.fixture
def CycDepClauses() -> list[Clause]:
    return [
        Clause(
            Atom(1, [Var(0), Var(1)]),
            [Atom(2, [Var(1), Var(2)]), Atom(3, [Var(0), Var(2)])],
            [],
            [],
        ),
        Clause(Atom(2, [Var(0), Var(1)]), [Atom(3, [Var(0), Var(1)])], [], []),
        Clause(Atom(3, [Var(0), Var(1)]), [], [Atom(1, [Var(1), Var(0)])], []),
    ]


def test_dependency_graph_cyclic(CycDepClauses):
    dg = evaluator._get_dependency_graph(
        5, [evaluator._normalize_clause(c, 0, 4) for c in CycDepClauses]
    )
    assert len(dg) == 5
    assert set(dg[0]) == set()
    assert set(dg[1]) == set([2, 3])
    assert set(dg[2]) == set([3])
    assert set(dg[3]) == set([1, 4])


def test_dependent_components_cyclic(CycDepClauses):
    dg = evaluator._get_dependency_graph(
        5, [evaluator._normalize_clause(c, 0, 4) for c in CycDepClauses]
    )
    components = evaluator._get_dependent_components(dg)
    assert len(components) == 3
    assert set(components[0]) == set([0])
    assert set(components[1]) == set([4])
    assert set(components[2]) == set([1, 2, 3])


def test_is_not_stratified(CycDepClauses):
    clauses = [evaluator._normalize_clause(c, 0, 4) for c in CycDepClauses]
    dg = evaluator._get_dependency_graph(5, clauses)
    components = evaluator._get_dependent_components(dg)
    assert not evaluator._is_stratified(5, clauses, components)


@pytest.fixture
def TakeProductRules() -> list[evaluator.NormalizedClause]:
    return list((
        evaluator._normalize_clause(clause, 0, 4)
        for clause in [
            Clause(
                Atom(1, [Var(0), Var(1)]),
                [Atom(2, [Var(0)]), Atom(3, [Var(1)])],
                [],
                [],
            )
        ]
    ))


def test_qt_generation_product(TakeProductRules):
    jg = evaluator.construct_join_graph(
        2, TakeProductRules[0].positive, TakeProductRules[0].negative
    )
    assert len(jg.nodes) == 2
    assert len(jg.arcs) == 2
    assert len(jg.arcs[0]) == 0
    assert len(jg.arcs[1]) == 0

    planner = evaluator.GreedyOptimizer(evaluator._cost_function)
    jt = planner(jg)
    assert jt.get_relations() == set([2, 3])
    assert set(jt.get_argument_map().keys()) == set([0, 1])
    # print(str(jt))


def test_qt_compilation_product(TakeProductRules):
    # code = evaluator.compile_without_dependencies(
    #     0, (0, 1), evaluator._generate_query_tree(TakeProductRules[0])
    # )
    compiled = evaluator._get_query_engine_code(5, TakeProductRules)
    a = set((tuple([i]) for i in range(0, 10, 2)))
    b = set((tuple([i]) for i in range(1, 10, 2)))
    env = {
        evaluator.RELATIONS: [set(), set(), a, b],
        evaluator.FLUENTS: [],
    }
    exec(compiled, env)
    result = env[evaluator.RELATIONS][1]
    assert result == set(itertools.product(range(0, 10, 2), range(1, 10, 2)))


@pytest.fixture
def JoinRules() -> list[evaluator.NormalizedClause]:
    return list((
        evaluator._normalize_clause(clause, 0, 4)
        for clause in [
            Clause(
                Atom(1, [Var(0), Var(1)]),
                [Atom(2, [Var(0), Var(2)]), Atom(3, [Var(2), Var(1)])],
                [],
                [],
            )
        ]
    ))


def test_qt_generation_join(JoinRules):
    jg = evaluator.construct_join_graph(3, JoinRules[0].positive, JoinRules[0].negative)
    assert len(jg.nodes) == 2
    assert len(jg.arcs) == 2
    assert len(jg.arcs[0]) == 1
    assert len(jg.arcs[1]) == 1

    planner = evaluator.GreedyOptimizer(evaluator._cost_function)
    jt = planner(jg)
    assert jt.get_relations() == set([2, 3])
    assert set(jt.get_argument_map().keys()) == set([0, 1, 2])
    # print(str(jt))


def test_qt_compilation_join(JoinRules):
    compiled = evaluator._get_query_engine_code(5, JoinRules)
    env = {
        evaluator.FLUENTS: [],
        evaluator.RELATIONS: [
            set(),
            set([(0, -1)]),
            set((i, i + 1) for i in range(1, 10, 1)),
            set((i + 1, i - 1) for i in range(1, 10, 1)),
        ],
    }
    exec(compiled, env)
    result = env[evaluator.RELATIONS][1]
    assert result == set(((i, i - 1) for i in range(10)))


@pytest.fixture
def JoinRulesWithDiff() -> list[evaluator.NormalizedClause]:
    return list((
        evaluator._normalize_clause(clause, 0, 5)
        for clause in [
            Clause(
                Atom(1, [Var(0), Var(1)]),
                [Atom(2, [Var(0), Var(2)]), Atom(3, [Var(2), Var(1)])],
                [Atom(4, [Var(2)])],
                [],
            )
        ]
    ))


def test_qt_compilation_join_with_diff(JoinRulesWithDiff):
    compiled = evaluator._get_query_engine_code(6, JoinRulesWithDiff)
    env = {
        evaluator.FLUENTS: [],
        evaluator.RELATIONS: [
            set(),
            set([(0, -1)]),
            set((i, i + 1) for i in range(1, 20, 1)),
            set((i + 1, i - 1) for i in range(1, 20, 1)),
            set((i,) for i in range(0, 21, 2)),
        ],
    }
    exec(compiled, env)
    result = env[evaluator.RELATIONS][1]
    assert result == set(((i, i - 1) for i in range(0, 20, 2)))


def test_qt_generation_join_with_diff(JoinRulesWithDiff):
    jg = evaluator.construct_join_graph(
        3, JoinRulesWithDiff[0].positive, JoinRulesWithDiff[0].negative
    )
    assert len(jg.nodes) == 3
    assert len(jg.arcs) == 3
    assert len(jg.arcs[0]) == 2
    assert len(jg.arcs[1]) == 2

    planner = evaluator.GreedyOptimizer(evaluator._cost_function)
    jt = planner(jg)
    # print(str(jt))
    assert jt.get_relations() == set([2, 3, 4])
    assert set(jt.get_argument_map().keys()) == set([0, 1, 2])


@pytest.fixture
def JoinWithFilter() -> list[evaluator.NormalizedClause]:
    return list((
        evaluator._normalize_clause(clause, 0, 5)
        for clause in [
            Clause(
                Atom(1, [Var(0), Var(1)]),
                [
                    Atom(2, [Var(0), Var(2)]),
                    Atom(3, [Var(1)]),
                    Atom(0, [Var(0), Obj(2)]),
                ],
                [Atom(4, [Var(2)]), Atom(0, [Var(1), Obj(2)])],
                [],
            )
        ]
    ))


def test_qt_generation_join_with_filters(JoinWithFilter):
    clause = JoinWithFilter[0]
    jg = evaluator.construct_join_graph(4, clause.positive, clause.negative)

    planner = evaluator.GreedyOptimizer(evaluator._cost_function)
    jt = planner(jg)
    # print(str(jt))
    jt = evaluator.insert_filter_predicates(
        clause.vars_eq, clause.vars_neq, clause.obj_eq, clause.obj_neq, jt
    )
    jt = evaluator.insert_projections(jt, set(clause.head.get_variables()))
    # print(str(jt))


def test_qt_compilation_joint_with_filters_empty(JoinWithFilter):
    compiled = evaluator._get_query_engine_code(6, JoinWithFilter)
    env = {
        evaluator.FLUENTS: [],
        evaluator.RELATIONS: [
            set(((i, i) for i in range(100))),
            set(),
            set(((i, i + 1) for i in range(99))),
            set(((i,) for i in range(0, 99, 2))),
            set(((i,) for i in range(1, 99, 2))),
        ],
    }
    exec(compiled, env)
    result = env[evaluator.RELATIONS][1]
    assert result == set()


def test_qt_compilation_joint_with_filters(JoinWithFilter):
    compiled = evaluator._get_query_engine_code(6, JoinWithFilter)
    env = {
        evaluator.FLUENTS: [],
        evaluator.RELATIONS: [
            set(((i, i) for i in range(100))),
            set(),
            set(((i, i + 1) for i in range(99))),
            set(((i,) for i in range(0, 99, 2))),
            set(((i,) for i in range(4, 99, 1))),
        ],
    }
    exec(compiled, env)
    result = env[evaluator.RELATIONS][1]
    assert result == set(((2, i) for i in range(0, 99, 2) if i != 2))


@pytest.fixture
def TransitiveClosure():
    return list((
        evaluator._normalize_clause(clause, 0, 3)
        for clause in [
            Clause(Atom(1, [Var(0), Var(1)]), [Atom(0, [Var(0), Var(1)])], [], []),
            Clause(
                Atom(1, [Var(0), Var(2)]),
                [Atom(2, [Var(1), Var(2)]), Atom(1, [Var(0), Var(1)])],
                [],
                [],
            ),
        ]
    ))


def test_evaluation_closure(TransitiveClosure):
    compiled = evaluator._get_query_engine_code(4, TransitiveClosure)
    env = {
        evaluator.FLUENTS: [],
        evaluator.RELATIONS: [
            set(((i, i) for i in range(100))),
            set(),
            set(((i, i + 1) for i in range(99))),
            set(((i,) for i in range(100)))
        ],
    }
    exec(compiled, env)
    result = env[evaluator.RELATIONS][1]
    assert result == set(((i, j) for i in range(100) for j in range(i, 100)))



from collections.abc import Iterable

from plado.pddl.arguments import ArgumentDefinition
from plado.pddl.boolean_expression import BooleanExpression
from plado.pddl.effects import ActionEffect


class Action:
    def __init__(
        self,
        name: str,
        parameters: Iterable[ArgumentDefinition],
        precondition: BooleanExpression,
        effect: ActionEffect,
    ):
        self.name: str = name
        self.parameters: tuple[ArgumentDefinition] = tuple(parameters)
        self.precondition: BooleanExpression = precondition
        self.effect: ActionEffect = effect

    def dump(self, level: int) -> str:
        return (
            f"(:action {self.name}\n{' ' * (level * 2)}  :parameters"
            f" ({' '.join((str(arg) for arg in self.parameters))})\n{' ' * (level * 2)} "
            f" :precondition {self.precondition.dump(level + 1)}\n{' ' * (level * 2)} "
            f" :effect {self.effect.dump(level + 1)}\n)"
        )



import itertools
from collections.abc import Iterable

from plado.datalog.program import Atom


class Node:
    def __init__(self, node_id: int, relation_id: int, args: tuple[int], negated: bool):
        self.node_id: int = node_id
        self.relation_id: int = relation_id
        self.args: tuple[int] = args
        self.negated: bool = negated

    def __repr__(self) -> str:
        return (
            f"JoinGraphNode({self.negated}{self.node_id}{self.relation_id}{self.args})"
        )

    def __hash__(self) -> int:
        return hash(self.node_id)

    def __eq__(self, other) -> bool:
        return isinstance(other, Node) and self.node_id == other.node_id


class JoinGraph:
    def __init__(self):
        self.nodes: list[Node] = None
        self.arcs: list[list[Node]] = None
        self.var_to_nodes: list[list[int]] = None


def create_node(atom: Atom, idx: int, negative: bool) -> Node:
    return Node(
        idx, atom.relation_id, tuple((arg.id for arg in atom.arguments)), negative
    )


def construct_join_graph(
    num_variables: int, positive: Iterable[Atom], negative: Iterable[Atom]
) -> JoinGraph:
    jg = JoinGraph()
    jg.nodes = [create_node(a, i, False) for i, a in enumerate(positive)]
    jg.nodes.extend(
        [create_node(a, i + len(jg.nodes), True) for i, a in enumerate(negative)]
    )
    jg.arcs = [[] for _ in range(len(jg.nodes))]
    jg.var_to_nodes = [[] for _ in range(num_variables)]
    # generate edges
    for i, source in enumerate(jg.nodes):
        # variable-to-node mappings
        for x in source.args:
            jg.var_to_nodes[x].append(i)
        # all nodes with a non-empty variable intersection -> the possible joins
        neighbors = sorted(
            set(
                itertools.chain.from_iterable((jg.var_to_nodes[x] for x in source.args))
            )
        )
        # create edge for each neighbor
        for j in neighbors:
            if i == j:
                continue
            target = jg.nodes[j]
            # negated edge (set minus) not commutative -> make case distinction
            # here (if not negated, insert edge in both directions):
            if not source.negated:
                jg.arcs[i].append(target)
            if not target.negated:
                jg.arcs[j].append(source)
    return jg



from plado import pddl
from plado.parser.parser import LookaheadStreamer, parse_domain, parse_problem
from plado.parser.sanity_checks import make_checks
from plado.parser.tokenizer import tokenize
from plado.pddl_utils.normalize import normalize_conditions, normalize_effects


def parse(
    domain_path: str, problem_path: str, skip_checks: bool = False
) -> tuple[pddl.Domain, pddl.Problem]:
    with open(domain_path, encoding="utf-8") as f:
        domain = parse_domain(LookaheadStreamer(tokenize(f.read())))
    with open(problem_path, encoding="utf-8") as f:
        problem = parse_problem(LookaheadStreamer(tokenize(f.read())))
    if not skip_checks:
        if make_checks(domain, problem):
            raise ValueError("invalid pddl syntax")
    return domain, problem


def parse_and_normalize(
    domain_path: str, problem_path: str, skip_checks: bool = False
) -> tuple[pddl.Domain, pddl.Problem]:
    domain, problem = parse(domain_path, problem_path, skip_checks)
    normalize_conditions(domain, problem)
    normalize_effects(domain)
    return domain, problem



from collections.abc import Iterable
from typing import Any

from plado.pddl.arguments import Argument, ArgumentDefinition, Substitution


class Function:
    def __init__(self, name: str, parameters: Iterable[ArgumentDefinition]):
        self.name: str = name
        self.parameters: tuple[ArgumentDefinition] = tuple(parameters)

    def __str__(self) -> str:
        return self.dump()

    def dump(self) -> str:
        return (
            f"({self.name}"
            f"{' ' if len(self.parameters) > 0 else ''}"
            f"{' '.join((str(arg) for arg in self.parameters))})"
        )


class NumericExpression:
    def traverse(self, visitor: "NumericExpressionVisitor") -> Any:
        raise NotImplementedError()

    def substitute(self, subs: Substitution) -> "NumericExpression":
        return self

    def get_free_variables(self) -> set[str]:
        raise NotImplementedError()

    def dump(self, level: int) -> str:
        raise NotImplementedError()


class FunctionCall(NumericExpression):
    def __init__(self, name: str, arguments: Iterable[Argument]):
        self.name: str = name
        self.arguments: tuple[Argument] = tuple(arguments)

    def traverse(self, visitor: "NumericExpressionVisitor") -> Any:
        return visitor.visit_function_call(self)

    def substitute(self, subs: Substitution) -> "FunctionCall":
        return FunctionCall(self.name, (arg.apply(subs) for arg in self.arguments))

    def dump(self, level: int) -> str:
        return (
            f"({self.name}"
            f"{' ' if len(self.arguments) > 0 else ''}"
            f"{' '.join((str(arg) for arg in self.arguments))}"
            ")"
        )

    def __str__(self):
        return "(%s%s)" % (self.name, " ".join((a.name for a in self.arguments)))

    def get_free_variables(self) -> set[str]:
        return set((arg.name for arg in self.arguments if arg.is_variable()))


class NumericConstant(NumericExpression):
    def __init__(self, value_str: str):
        self.value_str: str = value_str

    def traverse(self, visitor: "NumericExpressionVisitor") -> Any:
        return visitor.visit_constant(self)

    def get_free_variables(self) -> set[str]:
        return set()

    def dump(self, level: int) -> str:
        return self.value_str


class UnaryNegation(NumericExpression):
    def __init__(self, sub_expression: NumericExpression):
        self.sub_expression: NumericExpression = sub_expression

    def traverse(self, visitor: "NumericExpressionVisitor") -> Any:
        return visitor.visit_negation(visitor)

    def substitute(self, subs: Substitution) -> "UnaryNegation":
        return UnaryNegation(self.sub_expression.substitute(subs))

    def get_free_variables(self) -> set[str]:
        return self.sub_expression.get_free_variables()

    def dump(self, level: int) -> str:
        return f"(- {self.sub_expression.dump(level)})"


class BinaryOperation(NumericExpression):
    OP = None

    def __init__(self, lhs: NumericExpression, rhs: NumericExpression):
        self.lhs: NumericExpression = lhs
        self.rhs: NumericExpression = rhs

    def traverse(self, visitor: "NumericExpressionVisitor"):
        raise NotImplementedError()

    def substitute(self, subs: Substitution) -> "BinaryOperation":
        return self.__class__(self.lhs.substitute(subs), self.rhs.substitute(subs))

    def get_free_variables(self) -> set[str]:
        return self.lhs.get_free_variables() | self.rhs.get_free_variables()

    def dump(self, level: int) -> str:
        return f"({self.OP} {self.lhs.dump(level)} {self.rhs.dump(level)})"


class Sum(BinaryOperation):
    OP = "+"

    def traverse(self, visitor: "NumericExpressionVisitor") -> Any:
        return visitor.visit_sum(self)


class Subtraction(BinaryOperation):
    OP = "-"

    def traverse(self, visitor: "NumericExpressionVisitor") -> Any:
        return visitor.visit_subtraction(self)


class Product(BinaryOperation):
    OP = "*"

    def traverse(self, visitor: "NumericExpressionVisitor") -> Any:
        return visitor.visit_product(self)


class Division(BinaryOperation):
    OP = "/"

    def traverse(self, visitor: "NumericExpressionVisitor") -> Any:
        return visitor.visit_division(self)


class NumericExpressionVisitor:
    def visit_generic(self, expr: NumericExpression) -> Any:
        raise NotImplementedError()

    def visit_constant(self, expr: NumericConstant) -> Any:
        return self.visit_generic(expr)

    def visit_function_call(self, expr: FunctionCall) -> Any:
        return self.visit_generic(expr)

    def visit_negation(self, expr: UnaryNegation) -> Any:
        return self.visit_generic(expr)

    def visit_binary_op(self, expr: BinaryOperation) -> Any:
        return self.visit_generic(expr)

    def visit_sum(self, expr: Sum) -> Any:
        return self.visit_binary_op(expr)

    def visit_subtraction(self, expr: Subtraction) -> Any:
        return self.visit_binary_op(expr)

    def visit_product(self, expr: Product) -> Any:
        return self.visit_binary_op(expr)

    def visit_division(self, expr: Division) -> Any:
        return self.visit_binary_op(expr)


class RecursiveNumericExpressionVisitor(NumericExpressionVisitor):
    def visit_atomic(self, expr: NumericExpression) -> bool:
        raise NotImplementedError()

    def visit_constant(self, expr: NumericConstant) -> Any:
        return self.visit_atomic(expr)

    def visit_function_call(self, expr: FunctionCall) -> Any:
        return self.visit_atomic(expr)

    def visit_negation(self, expr: UnaryNegation) -> Any:
        return expr.sub_expression.traverse(self)

    def visit_binary_op(self, expr: NumericExpression) -> Any:
        if expr.lhs.traverse(self):
            return True
        return expr.rhs.traverse(self)


class NumericExpressionTransformer(NumericExpressionVisitor):
    def visit_generic(self, expr: NumericExpression) -> NumericExpression:
        return expr

    def visit_negation(self, expr: UnaryNegation) -> NumericExpression:
        return UnaryNegation(expr.sub_expression.traverse(self))

    def visit_binary_op(self, expr: BinaryOperation) -> NumericExpression:
        return expr.__class__(expr.lhs.traverse(self), expr.rhs.traverse(self))



from plado.pddl.numeric_expression import NumericExpression


class Metric:
    MINIMIZE = -1
    MAXIMIZE = 1

    def __init__(self, direction: int, expression: NumericExpression):
        self.direction: int = direction
        self.expression: NumericExpression = expression

    def dump(self, level: int = 0) -> str:
        return (
            f"(:{'minimize' if self.direction == Metric.MINIMIZE else 'maximize'} {self.expression.dump(level + 1)})"
        )



import itertools
from collections.abc import Callable, Iterable

import plado.datalog.numeric as datalog
from plado.datalog.evaluator.query_tree import (
    DifferenceNode,
    JoinNode,
    LeafNode,
    NumericConditionNode,
    PredicateNode,
    ProductNode,
    ProjectionNode,
    QNode,
    QueryTreeVisitor,
    RegisterNode,
    RelationArgs,
)
from plado.datalog.evaluator.query_tree_flattener import QueryTreeFlattener

MAX_LOOP_NESTING = 8

FLUENTS = "fluents"

RELATIONS = "relations"

VariableSubscripts = Callable[[int], str]


def get_indices(args: tuple[int], project: tuple[int]) -> tuple[int]:
    return tuple(args.index(i) for i in project)


def arg_to_position(args: tuple[int], relation_args: RelationArgs) -> tuple[int]:
    assert all((x in relation_args for x in args))
    return tuple((relation_args.index(x) for x in args))


class InstructionNode:

    def _indent(self, depth: int, *ins: str) -> str:
        def indent(depth: int, instr: str) -> str:
            return f'{" " * (2 * depth)}{instr}'

        return "\n".join((indent(depth, i) for i in ins))

    def to_string(self, depth: int = 0) -> str:
        raise NotImplementedError()


class MergeSet(InstructionNode):
    def __init__(self, dest: str, src: str):
        self.dest: str = dest
        self.src: str = src

    def to_string(self, depth: int = 0) -> str:
        return self._indent(depth, f"{self.dest}.update({self.src})")


class Assign(InstructionNode):
    def __init__(self, dest: str, src: str):
        self.dest: str = dest
        self.src: str = src

    def to_string(self, depth: int = 0) -> str:
        return self._indent(depth, f"{self.dest} = {self.src}")


class MakeSet(InstructionNode):
    def __init__(self, register: str):
        self.register: str = register

    def to_string(self, depth: int = 0) -> str:
        return self._indent(depth, f"{self.register} = set()")


class MakeDict(InstructionNode):
    def __init__(self, register: str):
        self.register: str = register

    def to_string(self, depth: int = 0) -> str:
        return self._indent(depth, f"{self.register} = {{}}")


class InstructionSequence(InstructionNode):
    def __init__(self, instructions: Iterable[InstructionNode]):
        self.instructions: tuple[InstructionNode] = tuple(instructions)

    def to_string(self, depth: int = 0) -> str:
        return "\n".join((i.to_string(depth) for i in self.instructions))


class ConditionalInstruction(InstructionNode):
    def __init__(self, instruction: InstructionNode, condition: str | None):
        self.instruction: InstructionNode = instruction
        self.condition: str | None = condition

    def to_string(self, depth: int = 0) -> str:
        if self.condition:
            return "\n".join([
                self._indent(depth, f"if {self.condition}:"),
                self.instruction.to_string(depth + 1),
            ])
        return self.instruction.to_string(depth)


class Insert(InstructionNode):
    def __init__(self, dest: str, val: str):
        self.dest: str = dest
        self.val: str = val

    def to_string(self, depth: int = 0) -> str:
        return self._indent(depth, f"{self.dest}.add({self.val})")


def InsertIf(dest: str, val: str, condition: str | None):
    return ConditionalInstruction(Insert(dest, val), condition)


class InsertDict(InstructionNode):
    def __init__(self, dest: str, key: str, val: str):
        self.dest: str = dest
        self.key: str = key
        self.val: str = val

    def to_string(self, depth: int = 0) -> str:
        a = self._indent(depth, f"if {self.key} not in {self.dest}:")
        b = self._indent(depth + 1, f"{self.dest}[{self.key}] = set()")
        c = self._indent(depth, f"{self.dest}[{self.key}].add({self.val})")
        return "\n".join((a, b, c))


def InsertDictIf(dest: str, key: str, val: str, condition: str | None):
    return ConditionalInstruction(InsertDict(dest, key, val), condition)


class ForLoop(InstructionNode):
    def __init__(
        self, var: str, container: str, condition: str | None, body: InstructionNode
    ):
        self.var: str = var
        self.container: str = container
        self.condition: str | None = condition
        self.body: InstructionNode = body

    def to_string(self, depth: int = 0) -> str:
        ins = []
        ins.append(self._indent(depth, f"for {self.var} in {self.container}:"))
        if self.condition:
            ins.append(self._indent(depth + 1, f"if not ({self.condition}):"))
            ins.append(self._indent(depth + 2, "continue"))
        ins.append(self.body.to_string(depth + 1))
        return "\n".join(ins)


class WhileLoop(InstructionNode):
    def __init__(self, condition: str, body: InstructionNode):
        self.condition: str = condition
        self.body: InstructionNode = body

    def to_string(self, depth: int = 0) -> str:
        return "\n".join([
            self._indent(depth, f"while {self.condition}:"),
            self.body.to_string(depth + 1),
        ])


class SetCompren(InstructionNode):
    def __init__(
        self, dest: str, val: str, var: str, container: str, condition: str | None
    ):
        self.dest: str = dest
        self.val: str = val
        self.container: str = container
        self.var: str = var
        self.condition: str | None = condition

    def to_string(self, depth: int = 0) -> str:
        ifstr = ""
        if self.condition:
            ifstr = f" if {self.condition}"
        return self._indent(
            depth,
            f"{self.dest} = set(({self.val} for {self.var} in"
            f" {self.container}{ifstr}))",
        )


class DictCompren(InstructionNode):
    def __init__(
        self,
        dest: str,
        key: str,
        val: str,
        var: str,
        container: str,
        condition: str | None,
    ):
        self.dest: str = dest
        self.key: str = key
        self.val: str = val
        self.container: str = container
        self.var: str = var
        self.condition: str | None = condition

    def to_string(self, depth: int = 0) -> str:
        instr = []
        instr.append(self._indent(depth, f"{self.dest} = {{}}"))
        instr.append(self._indent(depth, f"for {self.var} in {self.container}:"))
        if self.condition:
            instr.append(self._indent(depth + 1, f"if not ({self.condition}):"))
            instr.append(self._indent(depth + 2, "continue"))
        instr.append(self._indent(depth + 1, f"if {self.key} not in {self.dest}:"))
        instr.append(self._indent(depth + 2, f"{self.dest}[{self.key}] = set()"))
        instr.append(
            self._indent(depth + 1, f"{self.dest}[{self.key}].add({self.val})")
        )
        return "\n".join(instr)


class DelInstruction(InstructionNode):
    def __init__(self, register: str):
        self.register: str = register

    def to_string(self, depth: int = 0) -> str:
        return self._indent(depth, f"del {self.register}")


def get_fluent_arguments(fluent: datalog.Fluent, get_value: VariableSubscripts) -> str:
    args = [None] * len(fluent.args)
    for var, pos in fluent.variables:
        args[pos] = var
    return f"({', '.join((get_value(arg) for arg in args))},)"


def get_fluent(fluent_id: int, fluent_args: str) -> str:
    return f"{FLUENTS}[{fluent_id}].get({fluent_args}, None)"


AbstractDatabase = Callable[[int], str]


def are_disjoint(args0: tuple[int], args1: tuple[int]):
    s0 = sorted(args0)
    s1 = sorted(args1)
    i, j = 0, 0
    while i < len(s0) and j < len(s1):
        if s0[i] < s1[j]:
            i += 1
        elif s0[i] == s1[j]:
            return False
        else:
            j += 1
    return True


def get_var_pos_map(relation_args: RelationArgs) -> dict[int, int]:
    return {x: pos for (pos, x) in enumerate(relation_args)}


def project_tuple_to_indices(tuple_ref: str, indices: tuple[int]) -> str:
    if len(indices) == 0:
        return "tuple()"
    return f"({', '.join((f'{tuple_ref}[{i}]' for i in indices))},)"


def make_if_filter(condition: str | None) -> str:
    if condition is None:
        return ""
    return f" if {condition}"


class Predicate:
    def generate(self, values: VariableSubscripts) -> str:
        raise NotImplementedError()

    def __str__(self):
        return self.generate(lambda x: f"?x{x}")

    def is_equality_constraint(self) -> bool:
        return False

    def get_equality_constraints(self, dest: dict[int, int]):
        pass

    def get_variables(self) -> set[int]:
        raise NotImplementedError()


class NumericConstraint(Predicate):
    def __init__(self, constraint: datalog.NumericConstraint):
        self.constraint: datalog.NumericConstraint = constraint

    def get_variables(self) -> set[int]:
        return self.constraint.expr.get_variables()

    def generate(self, values: VariableSubscripts):
        class ASTExpr(datalog.NumericExpressionVisitor):
            def __init__(self):
                self.fluents: list[str] = []

            def visit_generic(self, expr: datalog.NumericExpression):
                raise ValueError()

            def visit_constant(self, expr: datalog.Constant) -> str:
                return str(expr.value)

            def visit_fluent(self, expr: datalog.Fluent) -> str:
                self.fluents.append(
                    get_fluent(expr.function_id, get_fluent_arguments(expr, values))
                )
                return self.fluents[-1]

            def visit_addition(self, expr: datalog.Addition) -> str:
                return f"{expr.lhs.accept(self)} + {expr.rhs.accept(self)}"

            def visit_subtraction(self, expr: datalog.Addition) -> str:
                return f"{expr.lhs.accept(self)} - {expr.rhs.accept(self)}"

            def visit_multiplication(self, expr: datalog.Addition) -> str:
                return f"({expr.lhs.accept(self)}) * ({expr.rhs.accept(self)})"

            def visit_division(self, expr: datalog.Addition) -> str:
                return f"({expr.lhs.accept(self)}) / ({expr.rhs.accept(self)})"

        visitor = ASTExpr()
        comparator = {
            datalog.NumericConstraint.GREATER: ">",
            datalog.NumericConstraint.GREATER_EQUAL: ">=",
            datalog.NumericConstraint.LESS: "<",
            datalog.NumericConstraint.LESS_EQUAL: "<=",
            datalog.NumericConstraint.EQUAL: "==",
        }[self.constraint.comparator]
        expr = self.constraint.expr.accept(visitor)
        constraint = f"{expr} {comparator} 0"
        not_none = (f"{fluent} is not None" for fluent in visitor.fluents)
        return " and ".join(itertools.chain(not_none, [constraint]))


class And(Predicate):
    def __init__(self, conjuncts: Iterable[Predicate]):
        self.conjuncts: list[Predicate] = []
        for pred in conjuncts:
            self.add(pred)

    def is_equality_constraint(self) -> bool:
        return all(c.is_equality_constraint() for c in self.conjuncts)

    def add(self, pred: Predicate):
        if isinstance(pred, And):
            self.conjuncts.extend(pred.conjuncts)
        else:
            self.conjuncts.append(pred)

    def generate(self, values: VariableSubscripts) -> str:
        return " and ".join([p.generate(values) for p in self.conjuncts])

    def get_variables(self) -> set[int]:
        return set(
            itertools.chain.from_iterable(
                itertools.chain((conj.get_variables() for conj in self.conjuncts))
            )
        )

    def get_equality_constraints(self, dest: dict[int, int]):
        for c in self.conjuncts:
            c.get_equality_constraints(dest)


class Equality(Predicate):
    Op = "=="

    def __init__(self, var0: int, var1: int):
        self.var0: int = var0
        self.var1: int = var1

    def generate(self, values: VariableSubscripts) -> str:
        return f"{values(self.var0)} {self.Op} {values(self.var1)}"

    def get_variables(self) -> set[int]:
        return set([self.var0, self.var1])


class Inequality(Equality):
    Op = "!="


class Select(Predicate):
    Op = "=="

    def __init__(self, var_id: int, constant: int):
        self.var: int = var_id
        self.constant: int = constant

    def generate(self, values: VariableSubscripts) -> str:
        return f"{values(self.var)} {self.Op} {self.constant}"

    def get_variables(self) -> set[int]:
        return set([self.var])

    def is_equality_constraint(self) -> bool:
        return True

    def get_equality_constraints(self, dest: dict[int, int]):
        dest[self.var] = self.constant


class SelectNot(Select):
    Op = "!="

    def is_equality_constraint(self) -> bool:
        return False

    def get_equality_constraints(self, dest: dict[int, int]):
        pass


def conjoin(*preds: Predicate) -> Predicate | None:
    conjoins = [p for p in preds if p is not None]
    return (
        None
        if len(conjoins) == 0
        else (conjoins[0] if len(conjoins) == 1 else And(conjoins))
    )


def make_condition(
    predicate: Predicate | None,
    tupl: str,
    relation_args: RelationArgs,
) -> str | None:
    if predicate is None:
        return None
    var_map = get_var_pos_map(relation_args)

    def get_var(var_id: int) -> str:
        assert var_id in var_map
        return f"{tupl}[{var_map[var_id]}]"

    return predicate.generate(get_var)


def make_condition_fun(
    predicate: Predicate | None, subscripts: VariableSubscripts
) -> str:
    if predicate is None:
        return None
    return predicate.generate(subscripts)


REGISTER = "_reg"


def get_register(num: int) -> str:
    return f"{REGISTER}{num}"


def _is_asc_sequence(seq: Iterable[int]) -> bool:
    last = None
    for i in seq:
        if last is None:
            last = i
        if last + 1 != i:
            return False
    return True


class Subscriptor:
    LOW = 0
    HIGH = 1

    def __init__(self):
        self.var_to_register_pos: dict[int, tuple[int, int, int]] = {}
        self.register_vars: dict[int, tuple[int]] = {}
        self.counter: int = 0

    def associate(self, register: int, args: tuple[int], priority: int) -> None:
        assert priority >= 0
        priority = self.counter
        self.counter += 1
        self.register_vars[register] = args
        for pos, arg in enumerate(args):
            if self.var_to_register_pos.get(arg, (None, None, -1))[2] < priority:
                self.var_to_register_pos[arg] = (register, pos, priority)

    def __call__(self, var_ref: int) -> str:
        assert var_ref in self.var_to_register_pos
        register, pos, _ = self.var_to_register_pos[var_ref]
        return f"{get_register(register)}[{pos}]"

    def get_tuple(self, args: Iterable[int]) -> str:
        args = list(args)
        if len(args) == 0:
            return "tuple()"
        assert all((arg in self.var_to_register_pos for arg in args))
        if len(set((self.var_to_register_pos[x][0]) for x in args)) == 1:
            reg, pos, _ = self.var_to_register_pos[args[0]]
            if _is_asc_sequence((self.var_to_register_pos[x][1] for x in args)):
                if len(args) == len(self.register_vars[reg]):
                    assert pos == 0
                    return get_register(reg)
                if pos + len(args) == len(self.register_vars[reg]):
                    return f"{get_register(reg)}[pos:]"
                return f"{get_register(reg)}[{pos}:{pos+len(args)}]"
        return f"({', '.join((self(x) for x in args))}, )"


TupleHandle = Callable[[VariableSubscripts], InstructionNode]


class Program:
    def __init__(self):
        self.num_registers: int = 0
        self.instructions: list[InstructionNode] = []

    def allocate_registers(self, num: int = 1) -> int:
        assert num >= 1
        self.num_registers += num
        return self.num_registers - num

    def free_registers(self, *register_ids: int) -> None:
        self.instructions.extend(
            (DelInstruction(get_register(r)) for r in register_ids)
        )

    def dump(self) -> str:
        return "\n".join((i.to_string(0) for i in self.instructions))


class BuildData:
    def __init__(self, dynamic_relations: Iterable[int]):
        self.dynamic_relations: set[int] = set(dynamic_relations)
        self.eval_cache = {}

    def clean_cache(self) -> list[InstructionNode]:
        instrs = []
        for cached in self.eval_cache.values():
            instrs.extend(cached.make_cleanup())
        return instrs


class PrebuiltHashSet:
    def make_is_not_in(self, handle: TupleHandle) -> TupleHandle:
        raise NotImplementedError()

    def make_loop_handle(
        self, register_id: int, subscripts: Subscriptor, handle: TupleHandle
    ) -> TupleHandle:
        raise NotImplementedError()

    def make_cleanup(self) -> list[InstructionNode]:
        raise NotImplementedError()


class PrebuiltHashSetRegister(PrebuiltHashSet):
    def __init__(self, register_id: int, relation_args: tuple[int]):
        self.register_id: int = register_id
        self.relation_args: tuple[int] = relation_args

    def make_is_not_in(self, handle: TupleHandle) -> TupleHandle:
        def handle_(subscripts: Subscriptor):
            return ConditionalInstruction(
                handle(subscripts),
                f"{subscripts.get_tuple(self.relation_args)} not in"
                f" {get_register(self.register_id)}",
            )

        return handle_

    def make_loop_handle(
        self, register_id: int, subscripts: Subscriptor, handle: TupleHandle
    ) -> TupleHandle:
        # subscripts.associate(register_id, self.relation_args, Subscriptor.HIGH)

        def handle_wrapper(subscripts: Subscriptor):
            subscripts.associate(register_id, self.relation_args, Subscriptor.HIGH)
            return ForLoop(
                get_register(register_id),
                get_register(self.register_id),
                None,
                handle(subscripts),
            )

        return handle_wrapper

    def make_cleanup(self) -> list[InstructionNode]:
        return [DelInstruction(get_register(self.register_id))]


class ConstantHashSetRelation(PrebuiltHashSet):
    def __init__(self, tupl: str, relation: str, relation_args: tuple[int]):
        self.tupl: str = tupl
        self.relation: str = relation
        self.relation_args: tuple[int] = relation_args

    def make_is_not_in(self, handle: TupleHandle) -> TupleHandle:
        def handle_wrapper(subscripts: Subscriptor):
            tupl: str = subscripts.get_tuple(self.relation_args)
            cond = f"{tupl} != {self.tupl}"
            return ConditionalInstruction(handle(subscripts), cond)

        return handle_wrapper

    def make_loop_handle(
        self, register_id: int, subscripts: Subscriptor, handle: TupleHandle
    ) -> TupleHandle:
        # subscripts.associate(register_id, self.relation_args, Subscriptor.HIGH)

        def handle_wrapper(subscripts: Subscriptor):
            subscripts.associate(register_id, self.relation_args, Subscriptor.HIGH)
            return ConditionalInstruction(
                InstructionSequence(
                    [Assign(get_register(register_id), self.tupl), handle(subscripts)]
                ),
                f"{self.tupl} in {self.relation}",
            )

        return handle_wrapper

    def make_cleanup(self) -> list[InstructionNode]:
        return []


class PrebuiltHashSetRelation(PrebuiltHashSet):
    def __init__(
        self, relation: str, relation_args: tuple[int], predicate: PredicateNode | None
    ):
        self.relation: str = relation
        self.relation_args: tuple[int] = relation_args
        self.predicate: PredicateNode | None = predicate

    def make_is_not_in(self, handle: TupleHandle) -> TupleHandle:
        def handle_wrapper(subscripts: Subscriptor):
            tupl: str = subscripts.get_tuple(self.relation_args)
            cond = f"{tupl} not in {self.relation}"
            if self.predicate:
                cond = (
                    f"{make_condition(self.predicate, tupl, self.relation_args)}"
                    f" and {cond}"
                )
            return ConditionalInstruction(handle(subscripts), cond)

        return handle_wrapper

    def make_loop_handle(
        self, register_id: int, subscripts: Subscriptor, handle: TupleHandle
    ) -> TupleHandle:
        # subscripts.associate(register_id, self.relation_args, Subscriptor.HIGH)

        def handle_wrapper(subscripts: Subscriptor):
            subscripts.associate(register_id, self.relation_args, Subscriptor.HIGH)
            return ForLoop(
                get_register(register_id),
                self.relation,
                make_condition(
                    self.predicate, get_register(register_id), self.relation_args
                ),
                handle(subscripts),
            )

        return handle_wrapper

    def make_cleanup(self) -> list[InstructionNode]:
        return []


class PrebuiltHashMap:
    def make_loop_handle(
        self, register_id: int, subscripts: Subscriptor, handle: TupleHandle
    ) -> TupleHandle:
        raise NotImplementedError()

    def make_cleanup(self) -> list[InstructionNode]:
        raise NotImplementedError()


class PrebuiltHashMapRegister(PrebuiltHashMap):
    def __init__(self, register_id: int, key_args: tuple[int], value_args: tuple[int]):
        self.register_id: int = register_id
        self.key_args: tuple[int] = key_args
        self.value_args: tuple[int] = value_args

    def make_loop_handle(
        self, register_id: int, subscripts: Subscriptor, handle: TupleHandle
    ) -> TupleHandle:
        # subscripts.associate(register_id, self.value_args, Subscriptor.LOW)

        def handle_wrapper(subscripts: Subscriptor):
            tupl: str = subscripts.get_tuple(self.key_args)
            subscripts.associate(register_id, self.value_args, Subscriptor.LOW)
            return ForLoop(
                get_register(register_id),
                f"{get_register(self.register_id)}.get({tupl}, [])",
                None,
                handle(subscripts),
            )

        return handle_wrapper

    def make_cleanup(self) -> list[InstructionNode]:
        return [DelInstruction(get_register(self.register_id))]


class PrebuiltHashMapRelation(PrebuiltHashMap):
    def __init__(
        self, relation: str, relation_args: tuple[int], predicate: PredicateNode | None
    ):
        self.relation: str = relation
        self.relation_args: tuple[int] = relation_args
        self.predicate: PredicateNode | None = predicate

    def make_loop_handle(
        self, register_id: int, subscripts: Subscriptor, handle: TupleHandle
    ) -> TupleHandle:
        def handle_wrapper(subscripts: Subscriptor):
            tupl: str = subscripts.get_tuple(self.relation_args)
            condition = f"{tupl} in {self.relation}"
            if self.predicate is not None:
                condition = (
                    f"{condition} and"
                    f" {make_condition(self.predicate, tupl, self.relation_args)}"
                )
            return ConditionalInstruction(handle(subscripts), condition)

        return handle_wrapper

    def make_cleanup(self) -> list[InstructionNode]:
        return []


class EvalBuilder:
    _INDEX = 0

    def __init__(self, predicate: Predicate | None, projection: tuple[int] | None):
        self._index = EvalBuilder._INDEX
        EvalBuilder._INDEX += 1
        self.predicate: Predicate | None = predicate
        self.projection: tuple[int] | None = projection

    def has_dynamic_dependency(self, dynamic_relations: set[int]) -> bool:
        raise NotImplementedError()

    def get_arguments(self) -> tuple[int]:
        raise NotImplementedError()

    def make_hash_index_set(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
        index_args: tuple[int],
    ) -> PrebuiltHashSet:
        raise NotImplementedError()

    def make_hash_index_map(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
        index_args: tuple[int],
    ) -> PrebuiltHashMap:
        raise NotImplementedError()

    def make_iteration(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
        handle: TupleHandle,
        subscripts: Subscriptor,
    ):
        raise NotImplementedError()

    def __hash__(self) -> int:
        return hash(self._index)

    def __eq__(self, o) -> bool:
        return self is o

    def _make_condition(self, subscripts: Subscriptor) -> str | None:
        return make_condition_fun(self.predicate, subscripts)

    def _dump(self) -> str:
        raise NotImplementedError()

    def __str__(self) -> str:
        return "".join([
            "[",
            self._dump(),
            f" where {str(self.predicate)}" if self.predicate else "",
            (
                f" project onto {sorted(self.projection)}"
                if self.projection is not None
                else ""
            ),
            "]",
        ])


class BinOpBuilder(EvalBuilder):
    OP = None

    def __init__(self, predicate: Predicate | None, projection: tuple[int] | None):
        super().__init__(predicate, projection)
        self.left: EvalBuilder | None = None
        self.right: EvalBuilder | None = None

    def _dump(self) -> str:
        return f"[{str(self.left)} {self.OP} {str(self.right)}]"

    def has_dynamic_dependency(self, dynamic_relations: set[int]) -> bool:
        return len(dynamic_relations) > 0 and (
            self.left.has_dynamic_dependency(dynamic_relations)
            or self.right.has_dynamic_dependency(dynamic_relations)
        )

    def store_child(self, child: EvalBuilder) -> bool:
        assert self.left is not None or self.right is None
        if self.left is None:
            self.left: EvalBuilder = child
            return False
        self.right: EvalBuilder = child
        return True

    def is_complete(self) -> bool:
        assert self.right is None or self.left is not None
        return self.right is not None

    def precompute_left_subtree(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
    ) -> PrebuiltHashSet:
        raise NotImplementedError()

    def precompute_right_subtree(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
    ) -> PrebuiltHashSet | PrebuiltHashMap:
        raise NotImplementedError()

    def combine_left_and_right_handle(
        self,
        program: Program,
        handle: TupleHandle,
        subscripts: Subscriptor,
        r_cache: PrebuiltHashSet | PrebuiltHashMap,
    ) -> TupleHandle:
        raise NotImplementedError()

    def iterate(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
        handle: TupleHandle,
        subscripts: Subscriptor,
    ):
        l_dyn: bool = self.left.has_dynamic_dependency(build_data.dynamic_relations)
        r_dyn: bool = self.right.has_dynamic_dependency(build_data.dynamic_relations)

        # 1. enumerate tuples and store them in allocated register
        # 1a. check if result is not already cached, compute result
        r_cache: PrebuiltHashMap | PrebuiltHashSet | None = build_data.eval_cache.get(
            self.right, None
        )

        # otherwise compute intermediate result
        if r_cache is None:
            r_cache = self.precompute_right_subtree(
                build_data,
                program,
                database,
            )

        # check if L has been cached
        l_cache: PrebuiltHashSet | None = build_data.eval_cache.get(self.left, None)

        # otherwise checker whether makes sense to store that intermediate
        # result persistently
        if l_cache is None and not l_dyn and r_dyn:
            l_cache: PrebuiltHashSet = self.precompute_left_subtree(
                build_data, program, database
            )
            # store in cache right away
            build_data.eval_cache[self.left] = l_cache

        # if L result not cached -> perform simple iteration
        if l_cache is None:
            # combine left and right sub trees
            self.left.make_iteration(
                build_data,
                program,
                database,
                self.combine_left_and_right_handle(
                    program, handle, subscripts, r_cache
                ),
                subscripts,
            )
        else:
            # need to "unpack" L tuple into tuples of individual
            # relations underlying L (handle code assumes L is always
            # iteratively constructed rather than precomputed)
            merged_tupl_rid: int = program.allocate_registers(1)
            # combine left and right sub trees
            handle = self.combine_left_and_right_handle(
                program, handle, subscripts, r_cache
            )
            program.instructions.append(
                l_cache.make_loop_handle(merged_tupl_rid, subscripts, handle)(
                    subscripts
                )
            )

        # cleanup in case necessary

        # l_cache is never cleaned up as, if computed, might be reused in next
        # iteration of fix point computation; r_cache might be reused only if
        # l_dyn is true while r_dyn is false ; so in that case, store in eval_cache (if not there
        # already); in all other cases cleanup r_cache
        if r_dyn or not l_dyn:
            program.instructions.extend(r_cache.make_cleanup())
            assert self.right not in build_data.eval_cache
        else:
            # store R cached result for next iteration
            build_data.eval_cache[self.right] = r_cache

    def make(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
        handle: TupleHandle,
    ):
        # create empty iterator
        subscripts: Subscriptor = Subscriptor()
        # construction
        self.iterate(build_data, program, database, handle, subscripts)

    def make_hash_index_set(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
        index_args: tuple[int],
    ) -> PrebuiltHashSet:
        # initialize registers
        register_id: int = program.allocate_registers(1)

        # program.instructions.append(MakeSet(get_register(register_id)))
        program.instructions.append(MakeSet(get_register(register_id)))

        # combine left and right sub-trees, projecting them to tuple to insert
        # into the hash set
        def combine(subscripts: Subscriptor) -> InstructionNode:
            # project onto index_args
            tupl: str = subscripts.get_tuple(self.get_arguments())
            # insert into set if predicate is satisfied
            return InsertIf(
                get_register(register_id), tupl, self._make_condition(subscripts)
            )

        # actual code generation
        self.make(build_data, program, database, combine)

        return PrebuiltHashSetRegister(register_id, index_args)

    def make_hash_index_map(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
        index_args: tuple[int],
    ) -> PrebuiltHashMap:
        # similar to make_hash_index_seast.Tuplet
        # initialize registers
        register_id: int = program.allocate_registers(1)
        # program.instructions.append(MakeDict(get_register(register_id)))

        program.instructions.append(MakeDict(get_register(register_id)))

        # combine left and right subtree
        def combine(subscripts: Subscriptor):
            # projection onto index_and value args
            key_tupl: str = subscripts.get_tuple(index_args)
            val_tupl: str = subscripts.get_tuple(self.get_arguments())
            # insert function:
            return InsertDictIf(
                get_register(register_id),
                key_tupl,
                val_tupl,
                self._make_condition(subscripts),
            )

        # actual code generation
        self.make(build_data, program, database, combine)

        return PrebuiltHashMapRegister(register_id, index_args, self.get_arguments())

    def make_iteration(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
        handle: TupleHandle,
        subscripts: Subscriptor,
    ):
        # similar to the other two make_ methods, except of not storing the
        # resulting tuples but calling handle instead
        # actual code generation
        # filter tuples with predicate before passing them onto handle
        def handle_wrapper(subscripts: Subscriptor):
            cond = self._make_condition(subscripts)
            return ConditionalInstruction(handle(subscripts), cond)

        # run code generation
        self.iterate(build_data, program, database, handle_wrapper, subscripts)


class DifferenceBuilder(BinOpBuilder):
    OP = "-"

    def __init__(
        self,
        shared_args: tuple[int],
        predicate: Predicate | None,
        projection: tuple[int] | None,
    ):
        super().__init__(predicate, projection)
        self.shared_args: tuple[int] = shared_args

    def get_arguments(self) -> tuple[int]:
        return self.left.get_arguments() if self.projection is None else self.projection

    def precompute_left_subtree(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
    ) -> PrebuiltHashSet:
        return self.left.make_hash_index_set(
            build_data, program, database, self.left.get_arguments()
        )

    def precompute_right_subtree(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
    ) -> PrebuiltHashSet | PrebuiltHashMap:
        return self.right.make_hash_index_set(
            build_data, program, database, self.shared_args
        )

    def combine_left_and_right_handle(
        self,
        program: Program,
        handle: TupleHandle,
        subscripts: Subscriptor,
        r_cache: PrebuiltHashSet,
    ) -> TupleHandle:
        return r_cache.make_is_not_in(handle)


class ProductBuilder(BinOpBuilder):
    OP = "x"

    def __init__(
        self,
        covered_args: tuple[int],
        predicate: Predicate | None,
        projection: tuple[int] | None,
    ):
        super().__init__(predicate, projection)
        self.covered_args: tuple[int] = covered_args

    def get_arguments(self) -> tuple[int]:
        return self.covered_args if self.projection is None else self.projection

    def precompute_left_subtree(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
    ) -> PrebuiltHashSet:
        return self.left.make_hash_index_set(
            build_data, program, database, self.left.get_arguments()
        )

    def precompute_right_subtree(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
    ) -> PrebuiltHashSet | PrebuiltHashMap:
        return self.right.make_hash_index_set(
            build_data, program, database, self.right.get_arguments()
        )

    def combine_left_and_right_handle(
        self,
        program: Program,
        handle: TupleHandle,
        subscripts: Subscriptor,
        r_cache: PrebuiltHashSet,
    ) -> TupleHandle:
        r_tupl_id: int = program.allocate_registers(1)
        return r_cache.make_loop_handle(r_tupl_id, subscripts, handle)


class JoinBuilder(ProductBuilder):
    OP = "."

    def __init__(
        self,
        join_args: tuple[int],
        covered_args: tuple[int],
        predicate: Predicate | None,
        projection: tuple[int] | None,
    ):
        super().__init__(covered_args, predicate, projection)
        self.join_args: tuple[int] = join_args

    def precompute_right_subtree(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
    ) -> PrebuiltHashSet | PrebuiltHashMap:
        return self.right.make_hash_index_map(
            build_data, program, database, self.join_args
        )

    def combine_left_and_right_handle(
        self,
        program: Program,
        handle: TupleHandle,
        subscripts: Subscriptor,
        r_cache: PrebuiltHashMap,
    ) -> TupleHandle:
        # register for iterating over right sub-tree tuples
        r_tupl_id: int = program.allocate_registers(1)
        return r_cache.make_loop_handle(r_tupl_id, subscripts, handle)


class TableScanBuilder(EvalBuilder):
    def __init__(
        self,
        relation_args: RelationArgs,
        predicate: Predicate | None = None,
        projection: tuple[int] | None = None,
    ):
        super().__init__(predicate, projection)
        self.relation_args: RelationArgs = relation_args

    def get_arguments(self) -> tuple[int]:
        return self.relation_args if self.projection is None else self.projection

    def _get_table(self, database: AbstractDatabase):
        raise NotImplementedError()

    def _is_constant(self, index_args: tuple[int]) -> bool:
        return (
            len(self.relation_args) > 0
            and (len(index_args) == 0 or len(index_args) == len(self.relation_args))
            and (
                self.predicate is not None
                and self.predicate.is_equality_constraint()
                and len(self.predicate.get_variables()) == len(self.relation_args)
            )
        )

    def _create_const_lookup(
        self, database: AbstractDatabase
    ) -> ConstantHashSetRelation:
        eq: dict[int, int] = {}
        self.predicate.get_equality_constraints(eq)
        tupl: str = (
            "(" + ", ".join((str(eq[x]) for x in self.relation_args)) + "," + ")"
        )
        return ConstantHashSetRelation(
            tupl, self._get_table(database), self.relation_args
        )

    def make_hash_index_set(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
        index_args: tuple[int],
    ) -> PrebuiltHashSet:
        if self._is_constant(index_args):
            return self._create_const_lookup(database)

        if set(index_args) == set(self.relation_args) and self.predicate is None:
            return PrebuiltHashSetRelation(
                self._get_table(database), self.relation_args, self.predicate
            )

        register_id: int = program.allocate_registers(1)
        tupl_id: int = program.allocate_registers(1)
        hash_indices = arg_to_position(index_args, self.relation_args)
        # inline assign set
        program.instructions.append(
            SetCompren(
                get_register(register_id),
                project_tuple_to_indices(get_register(tupl_id), hash_indices),
                get_register(tupl_id),
                self._get_table(database),
                make_condition(
                    self.predicate, get_register(tupl_id), self.relation_args
                ),
            )
        )
        # indicate that result stored in register
        return PrebuiltHashSetRegister(register_id, index_args)

    def make_hash_index_map(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
        index_args: tuple[int],
    ) -> PrebuiltHashMap:
        if set(index_args) == set(self.relation_args) and self.predicate is None:
            return PrebuiltHashMapRelation(
                self._get_table(database), self.relation_args, self.predicate
            )
        register_id: int = program.allocate_registers(1)
        # allocate temporary iterator variable
        tupl_id: int = program.allocate_registers(1)
        # project tuple onto those indices
        key_tupl = project_tuple_to_indices(
            get_register(tupl_id), get_indices(self.relation_args, index_args)
        )
        val_tupl = project_tuple_to_indices(
            get_register(tupl_id), get_indices(self.relation_args, self.get_arguments())
        )
        # for each tupl in the relation that satisfies the predicate,
        # add entry into hash table
        program.instructions.append(
            DictCompren(
                get_register(register_id),
                key_tupl,
                val_tupl,
                get_register(tupl_id),
                self._get_table(database),
                make_condition(
                    self.predicate, get_register(tupl_id), self.relation_args
                ),
            )
        )
        # indicate that result stored in register
        return PrebuiltHashMapRegister(register_id, index_args, self.get_arguments())

    def make_iteration(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
        handle: TupleHandle,
        subscripts: Subscriptor,
    ):
        # allocate iterator register
        register_id: int = program.allocate_registers(1)

        if self._is_constant(self.relation_args):
            cl = self._create_const_lookup(database)
            hndl = cl.make_loop_handle(register_id, subscripts, handle)
            program.instructions.append(hndl(subscripts))
            return

        # associate all relation arguments with the tuple stored in that
        # register
        subscripts.associate(register_id, self.relation_args, Subscriptor.HIGH)
        # inline assign set
        program.instructions.append(
            ForLoop(
                get_register(register_id),
                self._get_table(database),
                make_condition(
                    self.predicate, get_register(register_id), self.relation_args
                ),
                handle(subscripts),
            )
        )


class RelationScanBuilder(TableScanBuilder):
    def __init__(
        self,
        relation: int,
        relation_args: RelationArgs,
        predicate: Predicate | None,
        projection: tuple[int] | None,
    ):
        super().__init__(relation_args, predicate, projection)
        self.relation: int = relation

    def _dump(self) -> str:
        return f"P{self.relation}"

    def dump_tree(self) -> str:
        return f"P{self.relation}"

    def has_dynamic_dependency(self, dynamic_relations: set[int]) -> bool:
        return self.relation in dynamic_relations

    def _get_table(self, database: AbstractDatabase) -> str:
        return database(self.relation)


class RegisterScanBUilder(TableScanBuilder):
    def __init__(
        self,
        register: int,
        relation_args: RelationArgs,
    ):
        super().__init__(relation_args)
        self.register: int = register

    def _dump(self) -> str:
        return f"R{self.register}"

    def has_dynamic_dependency(self, dynamic_relations: set[int]) -> bool:
        return False

    def dump_tree(self) -> str:
        return f"R{self.register}"

    def _get_table(self, database: AbstractDatabase) -> str:
        return get_register(self.register)


class EmptyTupleBuilder(EvalBuilder):
    def __init__(self):
        super().__init__(None, None)

    def _dump(self) -> str:
        return "{()}"

    def has_dynamic_dependency(self, dynamic_relations: set[int]) -> bool:
        return False

    def get_arguments(self) -> tuple[int]:
        return tuple([])

    def make_hash_index_set(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
        index_args: tuple[int],
    ) -> PrebuiltHashSet:
        class HS(PrebuiltHashSet):
            def make_is_not_in(self, handle: TupleHandle) -> TupleHandle:
                def handle_wrapper(subscripts: Subscriptor):
                    return ConditionalInstruction(handle(subscripts), "False")

                return handle_wrapper

            def make_loop_handle(
                self, register_id: int, subscripts: Subscriptor, handle: TupleHandle
            ) -> TupleHandle:
                def handle_wrapper(subscripts: Subscriptor):
                    return ForLoop(
                        get_register(register_id), "[tuple()]", None, handle(subscripts)
                    )

                return handle_wrapper

            def make_cleanup(self) -> list[InstructionNode]:
                return []

        return HS()

    def make_hash_index_map(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
        index_args: tuple[int],
    ) -> PrebuiltHashMap:
        class HM(PrebuiltHashMap):
            def make_loop_handle(
                self, register_id: int, subscripts: Subscriptor, handle: TupleHandle
            ) -> TupleHandle:
                def handle_wrapper(subscripts: Subscriptor):
                    return ForLoop(
                        get_register(register_id), "[tuple()]", None, handle(subscripts)
                    )

                return handle_wrapper

            def make_cleanup(self) -> list[InstructionNode]:
                return []

        return HM()

    def make_iteration(
        self,
        build_data: BuildData,
        program: Program,
        database: AbstractDatabase,
        handle: TupleHandle,
        subscripts: Subscriptor,
    ):
        # allocate iterator register
        register_id: int = program.allocate_registers(1)
        program.instructions.append(
            ForLoop(
                get_register(register_id),
                "[tuple()]",
                None,
                handle(subscripts),
            )
        )


class ExecutionTreeBuilder(QueryTreeVisitor):
    """
    Traverses the join tree, collecting information needed for
    building up index structures to implement the joins.
    """

    def __init__(self):
        self.code_generators: list[EvalBuilder] = []
        self._build_stack: list[EvalBuilder] = []
        self._condition_stack: list[Predicate] = []
        self._projection: tuple[int] | None = None

    def visit_equality(self, node: PredicateNode) -> None:
        self._condition_stack.append(Equality(node.variable_id, node.value_ref))
        node.child.accept(self)

    def visit_select(self, node: PredicateNode) -> None:
        self._condition_stack.append(Select(node.variable_id, node.value_ref))
        node.child.accept(self)

    def visit_inequality(self, node: PredicateNode) -> None:
        self._condition_stack.append(Inequality(node.variable_id, node.value_ref))
        node.child.accept(self)

    def visit_select_not(self, node: PredicateNode) -> None:
        self._condition_stack.append(SelectNot(node.variable_id, node.value_ref))
        node.child.accept(self)

    def visit_projection(self, node: ProjectionNode) -> None:
        self._projection = self._projection or node.projection
        node.child.accept(self)

    def visit_numeric(self, node: NumericConditionNode) -> None:
        self._condition_stack.append(NumericConstraint(node.constraint))
        node.child.accept(self)

    def _consume(self) -> None:
        self._condition_stack.clear()
        self._projection = None

    def visit_product(self, node: ProductNode) -> None:
        self._build_stack.append(
            ProductBuilder(
                tuple(node.get_argument_map().keys()),
                conjoin(*self._condition_stack),
                self._projection,
            )
        )
        self._consume()
        node.left.accept(self)
        node.right.accept(self)

    def visit_join(self, node: JoinNode) -> None:
        self._build_stack.append(
            JoinBuilder(
                tuple(node.shared_args),
                tuple(node.get_argument_map().keys()),
                conjoin(*self._condition_stack),
                self._projection,
            )
        )
        self._consume()
        node.left.accept(self)
        node.right.accept(self)

    def visit_difference(self, node: DifferenceNode) -> None:
        self._build_stack.append(
            DifferenceBuilder(
                tuple(node.shared_args),
                conjoin(*self._condition_stack),
                self._projection,
            )
        )
        self._consume()
        node.left.accept(self)
        node.right.accept(self)

    def _backtrack(self, node: EvalBuilder):
        # insert node into parents
        while len(self._build_stack) > 0:
            if self._build_stack[-1].store_child(node):
                node = self._build_stack[-1]
                self._build_stack.pop(-1)
            else:
                break
        # if build stack is empty -> have processed entire execution
        if len(self._build_stack) == 0:
            self.code_generators.append(node)

    def visit_empty_tuple(self, node) -> None:
        node = EmptyTupleBuilder()
        self._consume()
        self._backtrack(node)

    def visit_leaf(self, node: LeafNode) -> None:
        # Table scan node
        node = RelationScanBuilder(
            node.relation_id,
            node.relation_args,
            conjoin(*self._condition_stack),
            self._projection,
        )
        self._consume()
        self._backtrack(node)

    def visit_register(self, node: RegisterNode) -> None:
        # Table scan node
        assert self._projection is None
        assert len(self._condition_stack) == 0
        node = RegisterNode(
            node.register_id,
            node.relation_args,
        )
        self._consume()
        self._backtrack(node)

    def visit_generic(self, node: QNode) -> None:
        raise ValueError(f"unexpected node of type {type(node)}")


def get_relation(relation: int) -> str:
    return f"{RELATIONS}[{relation}]"


def get_relation_delta(relation: int):
    return f"_delta{relation}"


def get_relation_delta_primed(relation: int):
    return f"_delta_prime{relation}"


def insert_into_handle(name: str, relation_args: list[int]):
    def handle(subscripts: Subscriptor):
        tupl: str = subscripts.get_tuple(relation_args)
        return Insert(name, tupl)

    return handle


def flatten_tree(
    tree: QNode, next_relation_id: int
) -> tuple[QNode, list[tuple[QNode, int]], int]:
    flattener = QueryTreeFlattener(MAX_LOOP_NESTING, next_relation_id)
    new_root = tree.accept(flattener)
    return (new_root, flattener.nodes, flattener.next_relation_id)


def compile_without_dependencies(
    relation: int,
    args: list[int],
    clause: QNode,
    num_relations: int,
) -> str:
    builder = ExecutionTreeBuilder()
    program = Program()

    def lookup_relation(relation: int) -> str:
        if relation < num_relations:
            return get_relation(relation)
        return f"_aux{relation - num_relations}"

    def insert_into_relation_handle(relation: int, relation_args: list[int]):
        return insert_into_handle(lookup_relation(relation), relation_args)

    # translate query trees into code gen trees
    root, temps, _ = flatten_tree(clause, num_relations)
    for node, _ in temps:
        node.accept(builder)
    root.accept(builder)
    assert len(builder.code_generators) == 1 + len(temps)

    # generate code
    build_data = BuildData([])

    # generate temporary results first (in order of appearence to satisfy
    # dependencies)
    for i, (node, target_relation) in enumerate(temps):
        gen = builder.code_generators[i]
        program.instructions.append(MakeSet(lookup_relation(target_relation)))
        gen.make_iteration(
            build_data,
            program,
            lookup_relation,
            insert_into_relation_handle(
                target_relation, tuple(sorted(node.get_argument_map().keys()))
            ),
            Subscriptor(),
        )
        # cleanup temporary results that are not needed anymore
        for relation_id in node.get_relations():
            if relation_id >= num_relations:
                program.instructions.append(
                    DelInstruction(lookup_relation(relation_id))
                )

    # generate final result
    builder.code_generators[len(temps)].make_iteration(
        build_data,
        program,
        lookup_relation,
        insert_into_relation_handle(relation, args),
        Subscriptor(),
    )
    # cleanup remaining temporary results
    for relation_id in root.get_relations():
        if relation_id >= num_relations:
            program.instructions.append(DelInstruction(lookup_relation(relation_id)))

    return InstructionSequence(program.instructions).to_string()


def compile_interdepending(
    relations: list[int],
    relations_args: list[tuple[int]],
    clauses: list[QNode],
    num_relations: int,
):
    assert len(relations) == len(relations_args) and len(relations_args) == len(clauses)
    builder = ExecutionTreeBuilder()
    build_data = BuildData(relations)

    def lookup_relation(relation: int) -> str:
        if relation < num_relations:
            return get_relation(relation)
        return f"_aux{relation - num_relations}"

    def insert_into_delta_prime_handle(relation: int, relation_args: tuple[int]):
        def handle(subscripts: Subscriptor):
            tupl: str = subscripts.get_tuple(relation_args)
            # insert into delta_prime if tupl not contained in delta and not in
            # relation
            return ConditionalInstruction(
                Insert(get_relation_delta_primed(relation), tupl),
                f"{tupl} not in {get_relation_delta(relation)}"
                f" and {tupl} not in {lookup_relation(relation)}",
            )

        return handle

    # flatten trees to bound maximal loop nesting and prepare code generation
    # trees
    flattened_relations: list[int] = []
    flattened_args: list[tuple[int]] = []
    next_relation_id: int = num_relations
    for node, relation_id, relation_args in zip(clauses, relations, relations_args):
        flattened, temps, next_relation_id = flatten_tree(node, next_relation_id)
        # prepare code generators for temporary results first
        for temp, temp_relation_id in temps:
            temp.accept(builder)
            if len(temp.get_relations() & build_data.dynamic_relations) > 0:
                build_data.dynamic_relations.add(temp_relation_id)
            flattened_relations.append(temp_relation_id)
            flattened_args.append(tuple(sorted(temp.get_argument_map().keys())))
        # prepare main code gen
        flattened.accept(builder)
        flattened_relations.append(relation_id)
        flattened_args.append(relation_args)
    assert len(builder.code_generators) == len(flattened_relations)

    relations = flattened_relations
    relations_args = flattened_args

    program = Program()

    # initialize temporary relations
    program.instructions.extend(
        (MakeSet(lookup_relation(r)) for r in range(num_relations, next_relation_id))
    )

    # initialize delta and delta'
    program.instructions.extend(
        [MakeSet(get_relation_delta(r)) for r in build_data.dynamic_relations]
    )
    program.instructions.extend(
        [MakeSet(get_relation_delta_primed(r)) for r in build_data.dynamic_relations]
    )

    # pre-compute what can be precomputed; count for each clause the number of
    # interdependent relations
    num_dependencies: list[int] = []

    def lookup_relation_and_count(relation_id: int):
        if relation_id in build_data.dynamic_relations:
            num_dependencies[-1] += 1
        return lookup_relation(relation_id)

    for relation, relation_args, build in zip(
        relations, relations_args, builder.code_generators
    ):
        num_dependencies.append(0)
        build.make_iteration(
            build_data,
            program,
            lookup_relation_and_count,
            insert_into_delta_prime_handle(relation, relation_args),
            Subscriptor(),
        )
    # generate inner loop code
    instructions, program.instructions = program.instructions, []
    # first step: merge delta with relations; move delta_prime into delta; reset
    # delta_prime
    program.instructions.extend([
        MergeSet(lookup_relation(r), get_relation_delta(r))
        for r in build_data.dynamic_relations
    ])
    program.instructions.extend([
        Assign(get_relation_delta(r), get_relation_delta_primed(r))
        for r in build_data.dynamic_relations
    ])
    program.instructions.extend(
        [MakeSet(get_relation_delta_primed(r)) for r in build_data.dynamic_relations]
    )

    class GetRelationOrDelta:
        def __init__(self, pos: int):
            self.counter: int = pos

        def __call__(self, relation_id: int):
            if relation_id not in build_data.dynamic_relations:
                return lookup_relation(relation_id)
            self.counter -= 1
            if self.counter == -1:
                return get_relation_delta(relation_id)
            return lookup_relation(relation_id)

    # generate inner loop join code:
    for relation, relation_args, build, num in zip(
        relations, relations_args, builder.code_generators, num_dependencies
    ):
        # generate num many rules, "swiping delta" along each such position
        for pos in range(num):
            build.make_iteration(
                build_data,
                program,
                GetRelationOrDelta(pos),
                insert_into_delta_prime_handle(relation, relation_args),
                Subscriptor(),
            )
    # outer loop
    # loop condition: while any delta_prime changed
    condition = " or ".join((
        f"len({get_relation_delta_primed(r)}) > 0" for r in build_data.dynamic_relations
    ))
    instructions.append(WhileLoop(condition, InstructionSequence(program.instructions)))
    instructions.extend([
        MergeSet(lookup_relation(r), get_relation_delta(r))
        for r in build_data.dynamic_relations
    ])
    instructions.extend(
        [DelInstruction(get_relation_delta(r)) for r in build_data.dynamic_relations]
    )
    instructions.extend([
        DelInstruction(get_relation_delta_primed(r))
        for r in build_data.dynamic_relations
    ])
    # cleanup cache
    instructions.extend(build_data.clean_cache())
    instructions.extend((
        DelInstruction(lookup_relation(r))
        for r in range(num_relations, next_relation_id)
    ))
    # done
    return InstructionSequence(instructions).to_string()



from typing import Any


class Type:
    def __init__(self, name: str, parent_type_name: str | None):
        self.name: str = name
        self.parent_type_name: str | None = parent_type_name

    def __hash__(self) -> int:
        return hash((self.__class__.__name__, self.name, self.parent_type_name))

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, Type)
            and other.name == self.name
            and other.parent_type_name == self.parent_type_name
        )

    def dump(self) -> str:
        return str(self)

    def __str__(self) -> str:
        if self.parent_type_name is not None:
            return f"{self.name} - {self.parent_type_name}"
        return self.name



from collections.abc import Iterable, Mapping
from typing import Any

from plado.utils import (
    Float,
    is_equal,
    is_greater,
    is_greater_equal,
    is_less,
    is_less_equal,
)

FluentsStore = list[dict[tuple[int], Float]]


class NumericExpression:
    def accept(self, visitor: "NumericExpressionVisitor") -> Any:
        raise NotImplementedError()

    def get_variables(self) -> set[int]:
        raise NotImplementedError()

    def substitute(self, sub: Mapping[int, int]) -> "NumericExpression":
        raise NotImplementedError()

    def evaluate(self, grounding: tuple[int], fluents: FluentsStore) -> Float:
        raise NotImplementedError()

    def __add__(self, other: "NumericExpression") -> "Addition":
        return Addition(self, other)

    def __sub__(self, other: "NumericExpression") -> "Subtraction":
        return Subtraction(self, other)

    def __mul__(self, other: "NumericExpression") -> "Multiplication":
        return Multiplication(self, other)

    def __div__(self, other: "NumericExpression") -> "Division":
        return Division(self, other)

    def __str__(self) -> str:
        raise NotImplementedError()


class Constant(NumericExpression):
    def __init__(self, value: Float):
        self.value: Float = value

    def get_variables(self) -> set[int]:
        return set()

    def accept(self, visitor: "NumericExpressionVisitor") -> Any:
        return visitor.visit_constant(self)

    def substitute(self, sub: Mapping[int, int]) -> "Constant":
        return self

    def evaluate(self, grounding: tuple[int], fluents: FluentsStore) -> Float:
        return self.value

    def __str__(self) -> str:
        return str(self.value)


class Fluent(NumericExpression):
    def __init__(
        self,
        function_id: int,
        args: Iterable[int],
        variables: Iterable[tuple[int, int]],
    ):
        self.function_id: int = function_id
        self.args: tuple[int] = tuple(args)
        self.variables: tuple[tuple[int, int]] = tuple(variables)

    def get_variables(self) -> set[int]:
        return set((x for x, _ in self.variables))

    def accept(self, visitor: "NumericExpressionVisitor") -> Any:
        return visitor.visit_fluent(self)

    def substitute(self, sub: Mapping[int, int]) -> "Fluent":
        return Fluent(
            self.function_id,
            self.args,
            ((sub[var], pos) for var, pos in self.variables),
        )

    def _ground_args(self, grounding: tuple[int]) -> tuple[int]:
        result = list(self.args)
        for var, pos in self.variables:
            result[pos] = grounding[var]
        return tuple(result)

    def evaluate(self, grounding: tuple[int], fluents: FluentsStore) -> Float:
        return fluents[self.function_id][self._ground_args(grounding)]

    def __str__(self) -> str:
        args = [str(arg) for arg in self.args]
        for var, pos in self.variables:
            args[pos] = f"?x{var}"
        sep = "" if len(self.args) == 0 else " "
        return f"(F{self.function_id}{sep}{' '.join(args)})"


class BinaryOperation(NumericExpression):
    OP = None

    def __init__(self, lhs: NumericExpression, rhs: NumericExpression):
        self.lhs: NumericExpression = lhs
        self.rhs: NumericExpression = rhs

    def get_variables(self) -> set[int]:
        return self.lhs.get_variables() | self.rhs.get_variables()

    def substitute(self, sub: Mapping[int, int]) -> "BinaryOperation":
        return self.__class__(
            self.lhs.substitute(sub),
            self.rhs.substitute(sub),
        )

    def __str__(self) -> str:
        return f"({self.OP} {self.lhs} {self.rhs})"


class Addition(BinaryOperation):
    OP = "+"

    def accept(self, visitor: "NumericExpressionVisitor") -> Any:
        return visitor.visit_addition(self)

    def evaluate(self, grounding: tuple[int], fluents: FluentsStore) -> Float:
        return self.lhs.evaluate(grounding, fluents) + self.rhs.evaluate(
            grounding, fluents
        )


class Multiplication(BinaryOperation):
    OP = "*"

    def accept(self, visitor: "NumericExpressionVisitor") -> Any:
        return visitor.visit_multiplication(self)

    def evaluate(self, grounding: tuple[int], fluents: FluentsStore) -> Float:
        return self.lhs.evaluate(grounding, fluents) * self.rhs.evaluate(
            grounding, fluents
        )


class Division(BinaryOperation):
    OP = "/"

    def accept(self, visitor: "NumericExpressionVisitor") -> Any:
        return visitor.visit_division(self)

    def evaluate(self, grounding: tuple[int], fluents: FluentsStore) -> Float:
        return self.lhs.evaluate(grounding, fluents) / self.rhs.evaluate(
            grounding, fluents
        )


class Subtraction(BinaryOperation):
    OP = "-"

    def accept(self, visitor: "NumericExpressionVisitor") -> Any:
        return visitor.visit_subtraction(self)

    def evaluate(self, grounding: tuple[int], fluents: FluentsStore) -> Float:
        return self.lhs.evaluate(grounding, fluents) - self.rhs.evaluate(
            grounding, fluents
        )


class NumericConstraint:
    """
    A numeric constraint of the form
    [expr] [op] 0
    where [expr] is any numeric expression, and [op] is in {<=, >=, ==}
    """

    LESS = -2
    LESS_EQUAL = -1
    EQUAL = 0
    GREATER_EQUAL = 1
    GREATER = 2

    def __init__(self, expr: NumericExpression, comparator: int):
        self.expr: NumericExpression = expr
        self.comparator: int = comparator

    def evaluate(self, grounding: tuple[int], fluents: FluentsStore) -> bool:
        lhs = self.expr.evaluate(grounding, fluents)
        if self.comparator == NumericConstraint.LESS_EQUAL:
            return is_less_equal(lhs, 0)
        if self.comparator == NumericConstraint.EQUAL:
            return is_equal(lhs, 0)
        if self.comparator == NumericConstraint.LESS:
            return is_less(lhs, 0)
        if self.comparator == NumericConstraint.GREATER:
            return is_greater(lhs, 0)
        return is_greater_equal(lhs, 0)

    def substitute(self, sub: Mapping[int, int]) -> "NumericConstraint":
        return NumericConstraint(self.expr.substitute(sub), self.comparator)

    def __str__(self) -> str:
        ops = {
            NumericConstraint.LESS: "<",
            NumericConstraint.LESS_EQUAL: "<=",
            NumericConstraint.GREATER: ">",
            NumericConstraint.GREATER_EQUAL: ">=",
            NumericConstraint.EQUAL: "==",
        }
        return f"({ops[self.comparator]} {self.expr} 0)"


class NumericExpressionVisitor:
    def visit_generic(self, expr: NumericExpression) -> Any:
        raise NotImplementedError()

    def visit_constant(self, expr: Constant) -> Any:
        return self.visit_generic(expr)

    def visit_fluent(self, expr: Fluent) -> Any:
        return self.visit_generic(expr)

    def visit_binary_operation(self, expr: BinaryOperation) -> Any:
        return self.visit_generic(expr)

    def visit_addition(self, expr: Addition) -> Any:
        return self.visit_binary_operation(expr)

    def visit_subtraction(self, expr: Subtraction) -> Any:
        return self.visit_binary_operation(expr)

    def visit_multiplication(self, expr: Multiplication) -> Any:
        return self.visit_binary_operation(expr)

    def visit_division(self, expr: Division) -> Any:
        return self.visit_binary_operation(expr)


def fluent_iterator(expr: NumericExpression) -> Iterable[Fluent]:
    class FluentVisitor(NumericExpressionVisitor):
        def visit_generic(self, expr: NumericExpression):
            yield from []

        def visit_binary_operation(self, expr: BinaryOperation) -> Iterable[Fluent]:
            yield from expr.lhs.accept(self)
            yield from expr.rhs.accept(self)

        def visit_fluent(self, expr: Fluent) -> Iterable[Fluent]:
            yield expr

    yield from expr.accept(FluentVisitor())



from typing import Any


def lower_bound(
        values: list[Any] | tuple[Any],
        value: Any,
        left: int,
        right: int,
        equal: lambda x, y: x == y,
        less: lambda x, y: x < y) -> int:
    '''
    Returns the first element in the range [left, right[ in values
    that does not compare less to value, or right if no such value exists.
    '''
    assert left >= 0
    assert right <= len(values)
    while left < right:
        mid = (left + right) / 2
        if equal(value, values[mid]) or less(value, values[mid]):
            right = mid
        else:
            left = mid + 1
    return right


def upper_bound(
        values: list[Any] | tuple[Any],
        value: Any,
        left: int,
        right: int,
        less: lambda x, y: x < y) -> int:
    '''
    Returns the first element in the range [left, right[ in values
    where value compares less to, or right if no such value exists.
    '''
    assert left >= 0
    assert right <= len(values)
    while left < right:
        mid = (left + right) / 2
        if less(value, values[mid]):
            right = mid
        else:
            left = mid + 1
    return left


def search(
        values: list[Any] | tuple[Any],
        value: Any,
        left: int,
        right: int,
        equal: lambda x, y: x == y,
        less: lambda x, y: x < y) -> int | None:
    '''
    Search in random-access container values for value in the range defined by
    [left, right[ (i.e., excluding right). Returns index of the element
    comparing equal to value, or None if no such element exists.
    '''
    pos = lower_bound(values, value, left, right, equal, less)
    if pos < right and equal(values[pos], value):
        return pos
    return None



from plado.datalog.evaluator.query_tree import (
    DifferenceNode,
    InnerNode,
    LeafNode,
    NumericConditionNode,
    PredicateNode,
    ProjectionNode,
    QNode,
    QueryTreeTransformer,
)


class QueryTreeSeparator(QueryTreeTransformer):
    def __init__(self, max_height: int, relation_id: int):
        self.max_height: int = max_height
        self.count: int = 0
        self.separated_node: QNode = None
        self.relation_id: int = relation_id
        self.has_predicate_parent: bool = False

    def visit_generic(self, node: QNode) -> QNode:
        self.count += 1
        return node

    def visit_difference(self, node: DifferenceNode) -> DifferenceNode:
        pp, self.has_predicate_parent = self.has_predicate_parent, True
        left = node.left.accept(self)
        if not pp and self.count == self.max_height:
            assert left is node.left
            self.separated_node = node
            return LeafNode(
                self.relation_id, tuple(sorted(node.get_argument_map().keys())), [], 0
            )
        if left is node.left:
            return node
        return DifferenceNode(left, node.right, node.cost)

    def visit_binary_op(self, node: InnerNode):
        pp, self.has_predicate_parent = self.has_predicate_parent, False
        left = node.left.accept(self)
        self.count += 1
        if not pp and self.count == self.max_height:
            assert left is node.left
            self.separated_node = node
            return LeafNode(
                self.relation_id, tuple(sorted(node.get_argument_map().keys())), [], 0
            )
        if left is node.left:
            return node
        return node.__class__(left, node.right, node.cost)

    def visit_predicate(self, node: PredicateNode) -> PredicateNode:
        pp, self.has_predicate_parent = self.has_predicate_parent, True
        child = node.child.accept(self)
        if not pp and self.count == self.max_height:
            assert child is node.child
            self.separated_node = node
            return LeafNode(
                self.relation_id, tuple(sorted(node.get_argument_map().keys())), [], 0
            )
        if child is node.child:
            return node
        return node.__class__(child, node.variable_id, node.value_ref)

    def visit_projection(self, node: ProjectionNode) -> ProjectionNode:
        pp, self.has_predicate_parent = self.has_predicate_parent, True
        child = node.child.accept(self)
        if not pp and self.count == self.max_height:
            assert child is node.child
            self.separated_node = node
            return LeafNode(
                self.relation_id, tuple(sorted(node.get_argument_map().keys())), [], 0
            )
        if node.child is child:
            return node
        return node.__class__(child, node.arg_map.keys())

    def visit_numeric(self, node: NumericConditionNode) -> NumericConditionNode:
        pp, self.has_predicate_parent = self.has_predicate_parent, True
        child = node.child.accept(self)
        if not pp and self.count == self.max_height:
            assert child is node.child
            self.separated_node = node
            return LeafNode(
                self.relation_id, tuple(sorted(node.get_argument_map().keys())), [], 0
            )
        if node.child is child:
            return node
        return node.__class__(child, node.constraint)


class QueryTreeFlattener(QueryTreeTransformer):
    def __init__(self, max_height: int, next_relation_id: int = 0):
        self.nodes: list[tuple[QNode, int]] = []
        self.next_relation_id: int = next_relation_id
        self.max_height: int = max_height
        assert self.max_height > 1

    def visit_difference(self, node: DifferenceNode) -> DifferenceNode:
        return DifferenceNode(
            node.left.accept(self), node.right.accept(self), node.cost
        )

    def _flatten(self, node: QNode) -> QNode:
        converged = False
        while not converged:
            converged = True
            sep = QueryTreeSeparator(self.max_height, self.next_relation_id)
            new_node = node.accept(sep)
            if sep.separated_node is not None:
                if sep.separated_node is node:
                    break
                converged = False
                self.nodes.append((sep.separated_node, self.next_relation_id))
                self.next_relation_id += 1
            node = new_node
        return node

    def visit_binary_op(self, node: InnerNode) -> QNode:
        right = self._flatten(node.right)
        new_node = node.__class__(node.left, right, node.cost)
        return self._flatten(new_node)



import argparse
import heapq
import itertools
import time
from collections.abc import Callable
from dataclasses import dataclass, field

from plado.parser import parse_and_normalize
from plado.semantics.applicable_actions_generator import ApplicableActionsGenerator
from plado.semantics.goal_checker import GoalChecker
from plado.semantics.successor_generator import SuccessorGenerator
from plado.semantics.task import State, Task

Heuristic = Callable[[State], int]


class Node:
    def __init__(
        self,
        state: State,
        parent: "Node",
        action: tuple[int, tuple[int]],
        g: int,
        h: int,
        status: int,
    ):
        self.state: State = state
        self.parent: Node | None = parent
        self.action: tuple[int, tuple[int]] = action
        self.g: int = g
        self.h: int = h
        self.status: int = status

    def mark_initial(self, h: int):
        self.g = 0
        self.h = h
        self.status = 0

    def open(self, parent: "Node", action: tuple[int, tuple[int]], g: int, h: int):
        self.parent = parent
        self.action = action
        self.g = g
        self.h = h
        self.status = 0

    def update_parent(self, parent: "Node", action: tuple[int, tuple[int]], g: int):
        self.parent = parent
        self.action = action
        self.g = g

    def close(self):
        self.status = 1

    def is_closed(self):
        return self.status == 1

    def is_new(self):
        return self.status == -1


@dataclass(order=True)
class PrioNode:
    f: tuple[int]
    node: Node = field(compare=False)


class Search:
    def __init__(self, task: Task, heuristic: Heuristic):
        self.task: Task = task
        self.aops_gen: ApplicableActionsGenerator = ApplicableActionsGenerator(task)
        self.succ_gen: SuccessorGenerator = SuccessorGenerator(task)
        self.is_goal: GoalChecker = GoalChecker(task)
        self.heuristic: Heuristic = heuristic
        self.search_space = {}
        self.queue: list[PrioNode] = []
        self.reopen_nodes: bool = False
        self.goal_node = None

    def _get_node(self, state: State) -> Node:
        key = tuple((tuple(sorted(atoms)) for atoms in state.atoms))
        return self.search_space.setdefault(key, Node(state, None, -1, -1, -1, -1))

    def _f(self, node: Node):
        return node.h, node.g

    def _push(self, node: Node):
        heapq.heappush(self.queue, PrioNode(self._f(node), node))

    def _push_initial(self, state: State):
        node = self._get_node(state)
        node.mark_initial(self.heuristic(state))
        self._push(node)

    def _pop(self) -> Node:
        if len(self.queue) == 0:
            raise StopIteration()
        pn = heapq.heappop(self.queue)
        return pn.node

    def _step(self) -> Node | None:
        node = self._pop()
        if self.is_goal(node.state):
            return node
        aops = self.aops_gen(node.state)
        for aop in aops:
            succs = self.succ_gen(node.state, aop)
            for succ, _ in succs:
                succ_node = self._get_node(succ)
                if succ_node.is_new():
                    succ_node.open(node, aop, node.g + 1, self.heuristic(succ))
                    self._push(succ_node)
                elif node.g + 1 < succ_node.g:
                    succ_node.update_parent(node, aop, node.g + 1)
                    if self.reopen_nodes or not node.is_closed():
                        self._push(succ_node)
        return None

    def _traceback(self, node: Node) -> list[str]:
        def ground_action(action: tuple[int, tuple[int]]) -> str:
            aid = action[0]
            params = action[1]
            return (
                f"({self.task.actions[aid].name}"
                f"{' ' if len(params) > 0 else ''}"
                f"{' '.join((self.task.objects[o] for o in params))})"
            )

        result: list[str] = []
        while node.parent is not None:
            result.append(ground_action(node.action))
            node = node.parent
        return list(reversed(result))

    def __call__(self, state: State) -> list[str] | None:
        self.search_space = {}
        self._push_initial(state)
        try:
            self.goal_node = None
            while self.goal_node is None:
                self.goal_node = self._step()
            return self._traceback(self.goal_node)
        except StopIteration:
            return None


class NoveltyHeuristic:
    def __init__(self, tuple_size: int):
        self.tuple_size: int = tuple_size
        self.db: set[tuple[int]] = set()

    def __call__(self, state: State) -> int:
        facts: list[tuple[int]] = []
        for p, args in enumerate(state.atoms):
            facts.append((p, *args))
        val = None
        for n in range(1, self.tuple_size):
            for combo in itertools.combinations(facts, n):
                t = tuple(itertools.chain(*combo))
                if val is None and t not in self.db:
                    val = n - 1
                self.db.add(t)
        if val is None:
            return self.tuple_size
        return val


class BlindHeuristic:
    def __call__(self, state: State) -> int:
        return 0


def main():
    p = argparse.ArgumentParser()
    p.add_argument("domain")
    p.add_argument("problem")
    p.add_argument("--novelty", "-w", type=int, default=2)
    p.add_argument("--verbose", "-v", action="store_true")
    p.add_argument(
        "--dump-and-exit",
        help="dump parsed and normalized pddl and exit",
        action="store_true",
    )
    args = p.parse_args()
    assert args.novelty >= 0

    t = time.time()

    if args.verbose:
        print("; parsing pddl...")
    domain, problem = parse_and_normalize(args.domain, args.problem)

    if args.dump_and_exit:
        print(domain.dump())
        print()
        print(problem.dump())
        return

    if args.verbose:
        print(f"; normalizing pddl... [t={time.time() - t}s]")
    task = Task(domain, problem)

    tt = time.time()
    if args.verbose:
        print(f"; starting search... [t={time.time() -t}s]")
    heuristic = (
        BlindHeuristic() if args.novelty == 0 else NoveltyHeuristic(args.novelty)
    )
    try:
        search = Search(task, heuristic)
        plan = search(task.initial_state)

        if plan is None:
            print("; unsolvable")
        else:
            if args.verbose:
                print("; plan found")
            print(f"; plan length: {len(plan)}")
            for i, func in enumerate(task.functions):
                if func.name == "total-cost":
                    print(f"; plan cost: {search.goal_node.state.fluents[i][tuple([])]}")
                if func.name == "reward":
                    print(f"; reward: {search.goal_node.state.fluents[i][tuple([])]}")
            print("\n".join(plan))
    except KeyboardInterrupt:
        print("interrupted")

    if args.verbose:
        print(f"; registered states: {len(search.search_space)}")
        print(f"; search time: {time.time() - tt}s")
        print(f"; total time: {time.time() - t}s")


if __name__ == "__main__":
    main()



from fractions import Fraction

Float = Fraction


def is_equal(x: Float, y: Float) -> bool:
    return x == y


def is_less_equal(x: Float, y: Float) -> bool:
    return x <= y


def is_greater_equal(x: Float, y: Float) -> bool:
    return x >= y


def is_less(x: Float, y: Float) -> bool:
    return x < y


def is_greater(x: Float, y: Float) -> bool:
    return x > y


def is_zero(number: Float) -> bool:
    return number == Fraction(0)
